__all__ = ["type_to_c"]

from functools import singledispatch

from ..ir.types import Array, FixedArray, Float, Integer, Mode, Pointer, Tensor, Type


def space_variable(variable: str | None = None) -> str:
    if variable is None:
        return ""
    else:
        return f" {variable}"


@singledispatch
def type_to_c(self: Type, variable: str | None = None) -> str:
    raise NotImplementedError(f"type_to_c not implemented for {type(self)}: {self}")


@type_to_c.register(Integer)
def type_to_c_integer(self: Integer, variable: str | None = None) -> str:
    return "int32_t" + space_variable(variable)


@type_to_c.register(Float)
def type_to_c_float(self: Float, variable: str | None = None) -> str:
    return "double" + space_variable(variable)


@type_to_c.register(Tensor)
def type_to_c_tensor(self: Tensor, variable: str | None = None) -> str:
    return "taco_tensor_t" + space_variable(variable)


@type_to_c.register(Mode)
def type_to_c_mode(self: Mode, variable: str | None = None) -> str:
    return "taco_mode_t" + space_variable(variable)


@type_to_c.register(Pointer)
def type_to_c_pointer(self: Pointer, variable: str | None = None) -> str:
    return f"{type_to_c(self.target)}* restrict" + space_variable(variable)


@type_to_c.register(Array)
def type_to_c_array(self: Array, variable: str | None = None) -> str:
    return f"{type_to_c(self.element, variable)}[]"


@type_to_c.register(FixedArray)
def type_to_c_fixed_array(self: FixedArray, variable: str | None = None) -> str:
    return f"{type_to_c(self.element, variable)}[{self.n}]"



__all__ = ["desugar_assignment"]

from functools import singledispatch
from itertools import count
from typing import Iterator

from ..expression import ast as sugar
from . import ast as desugar


@singledispatch
def desugar_expression(
    self: sugar.Expression, contract_indexes: set[str], ids: Iterator[int]
) -> desugar.Expression:
    raise NotImplementedError(f"desugar_expression not implemented for {type(self)}: {self}")


@desugar_expression.register(sugar.Integer)
def desugar_integer(
    self: sugar.Integer, contract_indexes: set[str], ids: Iterator[int]
) -> desugar.Expression:
    return desugar.Integer(self.value)


@desugar_expression.register(sugar.Float)
def desugar_float(
    self: sugar.Float, contract_indexes: set[str], ids: Iterator[int]
) -> desugar.Expression:
    return desugar.Float(self.value)


@desugar_expression.register(sugar.Tensor)
def desugar_tensor(
    self: sugar.Tensor, contract_indexes: set[str], ids: Iterator[int]
) -> desugar.Expression:
    output = desugar.Tensor(next(ids), self.name, self.indexes)
    for index in contract_indexes:
        output = desugar.Contract(index, output)
    return output


@desugar_expression.register(sugar.Add)
def desugar_add(
    self: sugar.Add, contract_indexes: set[str], ids: Iterator[int]
) -> desugar.Expression:
    left_indexes = set(self.left.index_participants().keys()).intersection(contract_indexes)
    right_indexes = set(self.right.index_participants().keys()).intersection(contract_indexes)

    intersection_indexes = left_indexes.intersection(right_indexes)

    output = desugar.Add(
        desugar_expression(self.left, left_indexes - intersection_indexes, ids),
        desugar_expression(self.right, right_indexes - intersection_indexes, ids),
    )

    for index in intersection_indexes:
        output = desugar.Contract(index, output)

    return output


@desugar_expression.register(sugar.Subtract)
def desugar_subtract(
    self: sugar.Subtract, contract_indexes: set[str], ids: Iterator[int]
) -> desugar.Expression:
    left_indexes = set(self.left.index_participants().keys()).intersection(contract_indexes)
    right_indexes = set(self.right.index_participants().keys()).intersection(contract_indexes)

    intersection_indexes = left_indexes.intersection(right_indexes)

    output = desugar.Add(
        desugar_expression(self.left, left_indexes - intersection_indexes, ids),
        desugar.Multiply(
            desugar.Integer(-1),
            desugar_expression(self.right, right_indexes - intersection_indexes, ids),
        ),
    )

    for index in intersection_indexes:
        output = desugar.Contract(index, output)

    return output


@desugar_expression.register(sugar.Multiply)
def desugar_multiply(
    self: sugar.Multiply, contract_indexes: set[str], ids: Iterator[int]
) -> desugar.Expression:
    left_indexes = set(self.left.index_participants().keys()).intersection(contract_indexes)
    right_indexes = set(self.right.index_participants().keys()).intersection(contract_indexes)

    intersection_indexes = left_indexes.intersection(right_indexes)

    output = desugar.Multiply(
        desugar_expression(self.left, left_indexes - intersection_indexes, ids),
        desugar_expression(self.right, right_indexes - intersection_indexes, ids),
    )

    for index in intersection_indexes:
        output = desugar.Contract(index, output)

    return output


def desugar_assignment(assignment: sugar.Assignment) -> desugar.Assignment:
    ids = count()

    desugared_target = desugar.Tensor(next(ids), assignment.target.name, assignment.target.indexes)

    all_indexes = set(assignment.index_participants().keys())
    contract_indexes = all_indexes - set(assignment.target.indexes)

    desugared_right_hand_side = desugar_expression(assignment.expression, contract_indexes, ids)

    return desugar.Assignment(desugared_target, desugared_right_hand_side)



__all__ = [
    "llvm_integer_type",
    "llvm_float_type",
    "llvm_boolean_type",
    "llvm_mode_type",
    "llvm_size_type",
    "type_to_llvm",
    "attribute_indexes",
]

from functools import singledispatch

import llvmlite.ir as llvm

from ..ir.types import Array, FixedArray, Float, Integer, Mode, Pointer, Tensor, Type

llvm_integer_type = llvm.IntType(32)
llvm_float_type = llvm.DoubleType()
llvm_boolean_type = llvm.IntType(1)
llvm_mode_type = llvm.IntType(8)
llvm_size_type = llvm.IntType(64)


@singledispatch
def type_to_llvm(self: Type) -> llvm.Type:
    raise NotImplementedError(f"type_to_llvm not implemented for {type(self)}: {self}")


@type_to_llvm.register(Integer)
def type_to_llvm_integer(self: Integer) -> llvm.Type:
    return llvm_integer_type


@type_to_llvm.register(Float)
def type_to_llvm_float(self: Float) -> llvm.Type:
    return llvm_float_type


tensor_attribute_indexes = {
    "dimensions": 1,
    "indices": 5,
    "vals": 6,
}


@type_to_llvm.register(Tensor)
def type_to_llvm_tensor(self: Tensor) -> llvm.Type:
    return llvm.LiteralStructType(
        [
            llvm_integer_type,  # order
            llvm.PointerType(llvm_integer_type),  # dimensions
            llvm_integer_type,  # csize
            llvm.PointerType(llvm_integer_type),  # mode_ordering
            llvm_mode_type,  # mode_types
            llvm.PointerType(llvm.PointerType(llvm.PointerType(llvm_integer_type))),  # indices
            llvm.PointerType(llvm_float_type),  # vals
            llvm_integer_type,  # vals_size
        ]
    )


@type_to_llvm.register(Mode)
def type_to_llvm_mode(self: Mode) -> llvm.Type:
    return llvm_mode_type


@type_to_llvm.register(Pointer)
def type_to_llvm_pointer(self: Pointer) -> llvm.Type:
    return llvm.PointerType(type_to_llvm(self.target))


@type_to_llvm.register(Array)
def type_to_llvm_array(self: Array) -> llvm.Type:
    return llvm.PointerType(type_to_llvm(self.element))


@type_to_llvm.register(FixedArray)
def type_to_llvm_fixed_array(self: FixedArray) -> llvm.Type:
    return llvm.ArrayType(type_to_llvm(self.element), self.n)


# LLVM index for all attributes.
# If we ever reuse an attribute name, and it goes to a different index, the whole IR will have to
# be redesigned.
attribute_indexes = tensor_attribute_indexes



__all__ = ["deparse_to_taco"]

from functools import singledispatch

from tensora.expression import ast


@singledispatch
def deparse_to_taco_expression(self: ast.Expression) -> str:
    raise NotImplementedError(
        f"deparse_to_taco_expression not implemented for {type(self)}: {self}"
    )


@deparse_to_taco_expression.register(ast.Integer)
def deparse_to_taco_integer(self: ast.Integer) -> str:
    return str(self.value)


@deparse_to_taco_expression.register(ast.Float)
def deparse_to_taco_float(self: ast.Float) -> str:
    return str(self.value)


@deparse_to_taco_expression.register(ast.Tensor)
def deparse_to_taco_tensor(self: ast.Tensor) -> str:
    if len(self.indexes) == 0:
        # Taco represents zero-dimensional tensors as scalars
        return self.name
    else:
        return f"{self.name}({', '.join(self.indexes)})"


@deparse_to_taco_expression.register(ast.Add)
def deparse_to_taco_add(self: ast.Add) -> str:
    return f"{deparse_to_taco_expression(self.left)} + {deparse_to_taco_expression(self.right)}"


@deparse_to_taco_expression.register(ast.Subtract)
def deparse_to_taco_subtract(self: ast.Subtract) -> str:
    return f"{deparse_to_taco_expression(self.left)} - {deparse_to_taco_expression(self.right)}"


@deparse_to_taco_expression.register(ast.Multiply)
def deparse_to_taco_multiply(self: ast.Multiply) -> str:
    left_string = deparse_to_taco_expression(self.left)
    if isinstance(self.left, (ast.Add, ast.Subtract)):
        left_string = f"({left_string})"

    right_string = deparse_to_taco_expression(self.right)
    if isinstance(self.right, (ast.Add, ast.Subtract)):
        right_string = f"({right_string})"

    return f"{left_string} * {right_string}"


def deparse_to_taco(self: ast.Assignment) -> str:
    return f"{deparse_to_taco_expression(self.target)} = {deparse_to_taco_expression(self.expression)}"



import pytest

from tensora.expression import (
    InconsistentDimensionsError,
    MutatingAssignmentError,
    NameConflictError,
    parse_assignment,
)
from tensora.expression.ast import *

assignment_strings = [
    (
        "A(i) = B(i,j) * C(j)",
        Assignment(Tensor("A", ("i",)), Multiply(Tensor("B", ("i", "j")), Tensor("C", ("j",)))),
    ),
    (
        "ab(i) = a(i) + b(i)",
        Assignment(Tensor("ab", ("i",)), Add(Tensor("a", ("i",)), Tensor("b", ("i",)))),
    ),
    (
        "D(i) = A(i) - B(i)",
        Assignment(Tensor("D", ("i",)), Subtract(Tensor("A", ("i",)), Tensor("B", ("i",)))),
    ),
    (
        "B2(i) = 2.0 * B(i)",
        Assignment(Tensor("B2", ("i",)), Multiply(Float(2.0), Tensor("B", ("i",)))),
    ),
    (
        "ab2(i) = 2.0 * (a(i) + b(i))",
        Assignment(
            Tensor("ab2", ("i",)),
            Multiply(Float(2.0), Add(Tensor("a", ("i",)), Tensor("b", ("i",)))),
        ),
    ),
    (
        "ab2(i) = (a(i) + b(i)) * 2.0",
        Assignment(
            Tensor("ab2", ("i",)),
            Multiply(Add(Tensor("a", ("i",)), Tensor("b", ("i",))), Float(2.0)),
        ),
    ),
]


@pytest.mark.parametrize(("string", "assignment"), assignment_strings)
def test_assignment_parsing(string, assignment):
    actual = parse_assignment(string).unwrap()
    assert actual == assignment


@pytest.mark.parametrize(("string", "assignment"), assignment_strings)
def test_assignment_deparsing(string, assignment):
    deparsed = assignment.deparse()
    assert deparsed == string


def test_mutating_assignment():
    assert isinstance(parse_assignment("A(i) = A(i) + 1").failure(), MutatingAssignmentError)


@pytest.mark.parametrize(
    "assignment",
    [
        "A(i) = B(i) + B(i,j)",
        "A(i) = B(i) - B()",
        "A(i) = B(i,j) * B(k,l,m)",
        "A(i) = B(i,j) + C(j,k) + (B(k) * D(k))",
    ],
)
def test_inconsistent_variable_size(assignment):
    assert isinstance(parse_assignment(assignment).failure(), InconsistentDimensionsError)


@pytest.mark.parametrize(
    "assignment",
    [
        "A(i) = B(B)",
        "A(i) = B(i,j) * C(j,B)",
        "A(i) = C(j,B) * B(i,j)",
        "A(A) = B(i)",
        "A(i) = B(A)",
        "A(B) = B(i)",
    ],
)
def test_name_conflict(assignment):
    assert isinstance(parse_assignment(assignment).failure(), NameConflictError)


def parse(string):
    return parse_assignment(string).unwrap()


def test_assignment_to_string():
    string = "A(i) = 2 * B(i,j) * (C(j) + D(j))"
    assert str(parse(string)) == string


@pytest.mark.parametrize(
    ("string", "output"),
    [
        (
            "y(i) = 0.5 * (b() - a()) * (x1(i,j) + x2(i,j)) * z(j)",
            {"y": 1, "b": 0, "a": 0, "x1": 2, "x2": 2, "z": 1},
        ),
        ("B2(i,k) = B(i,j) * B(j,k)", {"B2": 2, "B": 2}),
    ],
)
def test_variable_order(string, output):
    assert parse(string).variable_orders() == output


@pytest.mark.parametrize(
    ("string", "output"),
    [
        (
            "y(i) = 0.5 * (b() - a()) * (x1(i,j) + x2(i,j)) * z(j)",
            {"i": {("y", 0), ("x1", 0), ("x2", 0)}, "j": {("x1", 1), ("x2", 1), ("z", 0)}},
        ),
        (
            "B2(i,k) = B(i,j) * B(j,k)",
            {"i": {("B2", 0), ("B", 0)}, "j": {("B", 1), ("B", 0)}, "k": {("B2", 1), ("B", 1)}},
        ),
        ("diagA2(i) = A(i,i) + A(i,i)", {"i": {("diagA2", 0), ("A", 0), ("A", 1)}}),
    ],
)
def test_index_participants(string, output):
    assert parse(string).index_participants() == output



__all__ = [
    "TensorMethod",
    "BackendCompiler",
    "BroadcastTargetIndexError",
    "UnsupportedBackendCompilerError",
]

from dataclasses import dataclass
from enum import Enum
from inspect import Parameter, Signature

from returns.result import Failure, Success

from ..expression.ast import Assignment
from ..generate import Language, TensorCompiler, generate_code, generate_module_tensora
from ..kernel_type import KernelType
from ..problem import Problem
from ..tensor import Tensor
from ._cffi_ownership import allocate_taco_structure, take_ownership_of_arrays, tensor_cdefs


@dataclass(frozen=True, slots=True)
class BroadcastTargetIndexError(Exception):
    index: str
    assignment: Assignment

    def __str__(self):
        return (
            f"Expected index variable {self.index} on the target variable to be mentioned on the "
            f"right-hand side, but it was not: {self.assignment}. Such broadcasting makes sense "
            f"in a kernel and those kernels can be generated, but they cannot be used in "
            f"`evaluate` or `tensor_method` because those functions get the output dimensions "
            f"from the the dimensions of the input tensors."
        )


class BackendCompiler(Enum):
    """The tool to generate the machine code.

    Attributes
    ----------
    llvm
        Generate LLVM IR and compile with the llvmlite package.
        Not available with TensorCompiler.taco.
    cffi
        Generate C code and compile with the cffi package.
        Not available on Windows.
    """

    llvm = "llvm"
    cffi = "cffi"


@dataclass(frozen=True, slots=True)
class UnsupportedBackendCompilerError(Exception):
    backend_compiler: BackendCompiler
    tensor_compiler: TensorCompiler

    def __str__(self):
        return f"Backend compiler {self.backend_compiler} is not supported for tensor compiler {self.tensor_compiler}"


class TensorMethod:
    """A function taking specific tensor arguments."""

    def __init__(
        self,
        problem: Problem,
        compiler: TensorCompiler = TensorCompiler.tensora,
        backend: BackendCompiler = BackendCompiler.llvm,
    ):
        # Reject broadcasting to outputs because there is no way to specify output dimensions that
        # do not have a corresponding input dimension
        input_indexes = set(problem.assignment.expression.index_participants().keys())
        for output_index in problem.assignment.target.indexes:
            if output_index not in input_indexes:
                raise BroadcastTargetIndexError(output_index, problem.assignment)

        # Store validated attributes
        self._problem = problem
        self._output_name = problem.assignment.target.name
        self._input_formats = {
            name: format for name, format in problem.formats.items() if name != self._output_name
        }
        self._output_format = problem.formats[self._output_name]

        # Create Python signature of the function
        self.signature = Signature(
            [
                Parameter(parameter_name, Parameter.KEYWORD_ONLY, annotation=Tensor)
                for parameter_name in self._input_formats.keys()
            ]
        )

        match backend:
            case BackendCompiler.llvm:
                match compiler:
                    case TensorCompiler.tensora:
                        match generate_module_tensora(problem, [KernelType.evaluate]):
                            case Failure(error):
                                raise error
                            case Success(tensora_module):
                                from ._compile_llvm import compile_module

                                self._lib = compile_module(tensora_module)

                                # Convert ctypes function to cffi function
                                function_type = (
                                    f"int32_t (*)({', '.join(['void *'] * len(problem.formats))})"
                                )
                                function_pointer = self._lib.get_function_address("evaluate")
                                self._evaluate = tensor_cdefs.cast(function_type, function_pointer)
                    case TensorCompiler.taco:
                        raise UnsupportedBackendCompilerError(backend, compiler)
            case BackendCompiler.cffi:
                match generate_code(problem, [KernelType.evaluate], compiler, Language.c):
                    case Failure(error):
                        raise error
                    case Success(c_code):
                        from ._compile_cffi import compile_evaluate

                        self._lib = compile_evaluate(c_code)
                        self._evaluate = self._lib.evaluate

    def __call__(self, *args, **kwargs):
        # Handle arguments like normal Python function
        bound_arguments = self.signature.bind(*args, **kwargs).arguments

        # Validate tensor arguments
        for name, argument, format in zip(
            bound_arguments.keys(),
            bound_arguments.values(),
            self._input_formats.values(),
            strict=True,
        ):
            if not isinstance(argument, Tensor):
                raise TypeError(f"Argument {name} must be a Tensor not {type(argument)}")

            if argument.order != format.order:
                raise ValueError(
                    f"Argument {name} must have order {format.order} not {argument.order}"
                )
            if tuple(argument.modes) != tuple(format.modes):
                raise ValueError(
                    f"Argument {name} must have modes "
                    f"{tuple(mode.name for mode in format.modes)} not "
                    f"{tuple(mode.name for mode in argument.modes)}"
                )
            if tuple(argument.mode_ordering) != tuple(format.ordering):
                raise ValueError(
                    f"Argument {name} must have mode ordering "
                    f"{format.ordering} not {argument.mode_ordering}"
                )

        # Validate dimensions
        index_participants = self._problem.assignment.expression.index_participants()
        index_sizes = {}
        for index, participants in index_participants.items():
            # Extract the size of dimension referenced by this index on each tensor that uses it; record the variable
            # name and dimension for a better error
            actual_sizes = [
                (variable, dimension, bound_arguments[variable].dimensions[dimension])
                for variable, dimension in participants
            ]

            reference_size = actual_sizes[0][2]
            index_sizes[index] = reference_size

            for _, _, size in actual_sizes[1:]:
                if size != reference_size:
                    expected = ", ".join(
                        f"{variable}.dimensions[{dimension}] == {size}"
                        for variable, dimension, size in actual_sizes
                    )
                    raise ValueError(
                        f"{self._problem.assignment} expected all these dimensions of these tensors to be the same "
                        f"because they share the index {index}: {expected}"
                    )

        # Determine output dimensions
        output_dimensions = tuple(
            index_sizes[index] for index in self._problem.assignment.target.indexes
        )

        cffi_output = allocate_taco_structure(
            tuple(mode.c_int for mode in self._output_format.modes),
            output_dimensions,
            self._output_format.ordering,
        )

        output = Tensor(cffi_output)

        all_arguments = {self._output_name: output, **bound_arguments}

        cffi_args = [all_arguments[name].cffi_tensor for name in self._problem.formats.keys()]

        return_value = self._evaluate(*cffi_args)

        take_ownership_of_arrays(cffi_output)

        if return_value != 0:
            raise RuntimeError(f"Kernel evaluation failed with error code {return_value}")

        return output



import pytest

from tensora import Format, Mode, Tensor

try:
    # numpy is an optional dependency of tensora. It is only used to convert from numpy.
    import numpy
except ImportError:  # pragma: no cover
    numpy = None

try:
    # scipy is an optional dependency of tensora. It is only used to convert from scipy.sparse.
    import scipy.sparse as scipy_sparse
except ImportError:  # pragma: no cover
    scipy_sparse = None


def test_from_dok():
    dok = {
        (1, 2, 1): 4.5,
        (0, 2, 1): 2.0,
        (1, 0, 0): -2.0,
        (1, 2, 0): 4.5,
    }

    actual = Tensor.from_dok(dok, dimensions=(3, 3, 2), format="sss")

    assert actual.to_dok() == dok
    assert actual.format == Format((Mode.compressed, Mode.compressed, Mode.compressed), (0, 1, 2))
    assert actual.dimensions == (3, 3, 2)


def test_from_dok_default_dimensions_default_format():
    dok = {
        (1, 2, 1): 4.5,
        (0, 2, 1): 2.0,
        (1, 0, 0): -2.0,
        (1, 2, 0): 4.5,
    }

    actual_dok = Tensor.from_dok(dok)

    assert actual_dok.to_dok() == dok
    assert actual_dok.format == Format((Mode.dense, Mode.compressed, Mode.compressed), (0, 1, 2))
    assert actual_dok.dimensions == (2, 3, 2)


def test_from_soa():
    dok = {
        (1, 2, 1): 4.5,
        (0, 2, 1): 2.0,
        (1, 0, 0): -2.0,
        (1, 2, 0): 4.5,
    }

    soa = ([1, 0, 1, 1], [2, 2, 0, 2], [1, 1, 0, 0])

    values = [
        4.5,
        2.0,
        -2.0,
        4.5,
    ]

    actual = Tensor.from_soa(soa, values, dimensions=(3, 3, 2), format="sss")

    assert actual.to_dok() == dok
    assert actual.format == Format((Mode.compressed, Mode.compressed, Mode.compressed), (0, 1, 2))
    assert actual.dimensions == (3, 3, 2)


def test_from_soa_default_dimensions_default_format():
    dok = {
        (1, 2, 1): 4.5,
        (0, 2, 1): 2.0,
        (1, 0, 0): -2.0,
        (1, 2, 0): 4.5,
    }

    soa = ([1, 0, 1, 1], [2, 2, 0, 2], [1, 1, 0, 0])

    values = [
        4.5,
        2.0,
        -2.0,
        4.5,
    ]

    actual = Tensor.from_soa(soa, values)

    assert actual.to_dok() == dok
    assert actual.format == Format((Mode.dense, Mode.compressed, Mode.compressed), (0, 1, 2))
    assert actual.dimensions == (2, 3, 2)


def test_from_aos():
    dok = {
        (1, 2, 1): 4.5,
        (0, 2, 1): 2.0,
        (1, 0, 0): -2.0,
        (1, 2, 0): 4.5,
    }

    aos = [
        (1, 2, 1),
        (0, 2, 1),
        (1, 0, 0),
        (1, 2, 0),
    ]

    values = [
        4.5,
        2.0,
        -2.0,
        4.5,
    ]

    actual = Tensor.from_aos(aos, values, dimensions=(3, 3, 2), format="sss")

    assert actual.to_dok() == dok
    assert actual.format == Format((Mode.compressed, Mode.compressed, Mode.compressed), (0, 1, 2))
    assert actual.dimensions == (3, 3, 2)


def test_from_aos_default_dimensions_default_format():
    dok = {
        (1, 2, 1): 4.5,
        (0, 2, 1): 2.0,
        (1, 0, 0): -2.0,
        (1, 2, 0): 4.5,
    }

    aos = [
        (1, 2, 1),
        (0, 2, 1),
        (1, 0, 0),
        (1, 2, 0),
    ]

    values = [
        4.5,
        2.0,
        -2.0,
        4.5,
    ]

    actual = Tensor.from_aos(aos, values)

    assert actual.to_dok() == dok
    assert actual.format == Format((Mode.dense, Mode.compressed, Mode.compressed), (0, 1, 2))
    assert actual.dimensions == (2, 3, 2)


def test_from_lol():
    dok = {
        (1, 2, 1): 4.5,
        (0, 2, 1): 2.0,
        (1, 0, 0): -2.0,
        (1, 2, 0): 4.5,
    }

    lol = [[[0, 0], [0, 0], [0, 2]], [[-2, 0], [0, 0], [4.5, 4.5]], [[0, 0], [0, 0], [0, 0]]]

    actual = Tensor.from_lol(lol, dimensions=(3, 3, 2), format="sss")

    assert actual.to_dok() == dok
    assert actual.format == Format((Mode.compressed, Mode.compressed, Mode.compressed), (0, 1, 2))
    assert actual.dimensions == (3, 3, 2)


def test_from_lol_default_dimensions_default_format():
    dok = {
        (1, 2, 1): 4.5,
        (0, 2, 1): 2.0,
        (1, 0, 0): -2.0,
        (1, 2, 0): 4.5,
    }

    lol = [[[0, 0], [0, 0], [0, 2]], [[-2, 0], [0, 0], [4.5, 4.5]], [[0, 0], [0, 0], [0, 0]]]

    actual = Tensor.from_lol(lol)

    assert actual.to_dok() == dok
    assert actual.format == Format((Mode.dense, Mode.dense, Mode.dense), (0, 1, 2))
    assert actual.dimensions == (3, 3, 2)


def test_from_lol_scalar():
    actual = Tensor.from_lol(3.5)

    assert actual.to_dok() == {(): 3.5}
    assert actual.format == Format((), ())
    assert actual.dimensions == ()


def assert_all_methods_same(lol, dimensions, format):
    dok = {}

    def recurse(tree, indexes):
        if isinstance(tree, (int, float)):
            dok[indexes] = tree
        else:
            for i_element, element in enumerate(tree):
                recurse(element, (*indexes, i_element))

    recurse(lol, ())

    dok_no_zeros = {key: value for key, value in dok.items() if value != 0}
    coordinates = list(dok.keys())
    values = list(dok.values())
    coordinates_no_zeros = list(dok_no_zeros.keys())
    values_no_zeros = list(dok_no_zeros.values())

    tensor_from_lol = Tensor.from_lol(lol, dimensions=dimensions, format=format)
    tensor_from_dok = Tensor.from_dok(dok, dimensions=dimensions, format=format)
    tensor_from_dok_no_zeros = Tensor.from_dok(dok_no_zeros, dimensions=dimensions, format=format)
    tensor_from_aos = Tensor.from_aos(coordinates, values, dimensions=dimensions, format=format)
    tensor_from_aos_no_zeros = Tensor.from_aos(
        coordinates_no_zeros, values_no_zeros, dimensions=dimensions, format=format
    )

    assert tensor_from_dok == tensor_from_lol
    assert tensor_from_dok_no_zeros == tensor_from_lol
    assert tensor_from_aos == tensor_from_lol
    assert tensor_from_aos_no_zeros == tensor_from_lol

    if numpy is not None:
        # Apply reshape so that zero dimensions appear correctly
        numpy_array = numpy.asarray(lol).reshape(dimensions)
        tensor_from_numpy = Tensor.from_numpy(numpy_array, format=format)

        assert tensor_from_numpy == tensor_from_lol

    if scipy_sparse is not None and len(dimensions) == 2:
        # Scipy sparse only allows matrices
        if len(values) == 0:
            # zip only transposes if the iterable is not empty
            soa = ([], [])
        else:
            soa = tuple(zip(*coordinates, strict=False))
        scipy_csc_matrix = scipy_sparse.csc_matrix((values, soa), shape=dimensions)
        scipy_csr_matrix = scipy_sparse.csr_matrix((values, soa), shape=dimensions)
        scipy_coo_matrix = scipy_sparse.coo_matrix((values, soa), shape=dimensions)

        assert Tensor.from_scipy_sparse(scipy_csc_matrix) == tensor_from_lol
        assert Tensor.from_scipy_sparse(scipy_csr_matrix) == tensor_from_lol
        assert Tensor.from_scipy_sparse(scipy_coo_matrix) == tensor_from_lol


@pytest.mark.parametrize("lol", [3.5, 0.0, 1])
def test_convert_0(lol):
    assert_all_methods_same(lol, (), "")


@pytest.mark.parametrize(
    ("lol", "dimensions"),
    [
        ([], (0,)),
        ([3], (1,)),
        ([3, -2, 0], (3,)),
    ],
)
@pytest.mark.parametrize("format", ["d", "s"])
def test_convert_1(lol, dimensions, format):
    assert_all_methods_same(lol, dimensions, format)


@pytest.mark.parametrize(
    ("lol", "dimensions"),
    [
        ([], (0, 0)),
        ([], (0, 3)),
        ([[], []], (2, 0)),
        ([[0, -2, 3]], (1, 3)),
        ([[0], [3]], (2, 1)),
        ([[0, 2, 3], [0, -2, 3]], (2, 3)),
        ([[0, 0, 0], [0, 0, 0]], (2, 3)),
    ],
)
@pytest.mark.parametrize("format", ["ss", "dd", "sd", "ds", "s1s0", "d1d0", "s1d0", "d1s0"])
def test_convert_2(lol, dimensions, format):
    assert_all_methods_same(lol, dimensions, format)


@pytest.mark.parametrize(
    ("lol", "dimensions"),
    [
        ([], (0, 0, 0)),
        ([], (0, 2, 3)),
        ([[], [], [], []], (4, 0, 3)),
        ([[[], []], [[], []], [[], []], [[], []]], (4, 2, 0)),
        ([[[0, 0, 0], [1, 2, 3]]], (1, 2, 3)),
        ([[[0, 0, 0]], [[0, 0, 0]], [[0, 0, 1]], [[1, 0, 0]]], (4, 1, 3)),
        ([[[0], [1]], [[0], [-2]], [[0], [0]], [[0], [1]]], (4, 2, 1)),
        (
            [
                [[0, 0, 0], [1, 2, 3]],
                [[0, 0, 0], [-2, 3, 0]],
                [[0, 0, 0], [0, 0, 1]],
                [[0, 0, 0], [1, 0, 0]],
            ],
            (4, 2, 3),
        ),
        (
            [
                [[0, 0, 0], [0, 0, 0]],
                [[0, 0, 0], [0, 0, 0]],
                [[0, 0, 0], [0, 0, 0]],
                [[0, 0, 0], [0, 0, 0]],
            ],
            (4, 2, 3),
        ),
    ],
)
@pytest.mark.parametrize("modes", ["sss", "ssd", "sds", "sdd", "dss", "dsd", "dds", "ddd"])
@pytest.mark.parametrize("ordering", ["012", "021", "120", "102", "201", "210"])
def test_convert_3(lol, dimensions, modes, ordering):
    format = modes[0] + ordering[0] + modes[1] + ordering[1] + modes[2] + ordering[2]
    assert_all_methods_same(lol, dimensions, format)


@pytest.mark.parametrize(
    "data",
    [
        0,
        2.0,
    ],
)
def test_interconversion_0(data):
    expected = Tensor.from_lol(data)
    actual = expected.to_format("")

    assert expected == actual


@pytest.mark.parametrize(
    "data",
    [
        [0, 0],
        [0, 1],
    ],
)
@pytest.mark.parametrize("format1", ["d", "s"])
@pytest.mark.parametrize("format2", ["d", "s"])
def test_interconversion_1(data, format1, format2):
    expected = Tensor.from_lol(data)
    actual = expected.to_format(format2)

    assert expected == actual


@pytest.mark.parametrize(
    "data",
    [
        [[0, 0], [0, 0]],
        [[0, 1], [-2, 0]],
    ],
)
@pytest.mark.parametrize("format1", ["ss", "dd", "sd", "ds", "s1s0", "d1d0", "s1d0", "d1s0"])
@pytest.mark.parametrize("format2", ["ss", "dd", "sd", "ds", "s1s0", "d1d0", "s1d0", "d1s0"])
def test_interconversion_2(data, format1, format2):
    expected = Tensor.from_lol(data)
    actual = expected.to_format(format2)

    assert expected == actual



__all__ = ["InvalidModeOrderingError"]

from dataclasses import dataclass

from ._format import Mode


@dataclass(frozen=True, slots=True)
class InvalidModeOrderingError(Exception):
    modes: tuple[Mode, ...]
    ordering: tuple[int, ...]

    def __str__(self):
        return (
            f"Expected ordering to have be some order of integers 0 until length of modes, "
            f"but got modes={self.modes} and ordering={self.ordering}"
        )



from __future__ import annotations

__all__ = ["Tensor"]

import itertools
from numbers import Real
from typing import Any, Dict, Iterable, Iterator, List, Optional, Sequence, Tuple, Union

from .compile import taco_structure_to_cffi
from .format import Format, Mode, parse_format


class Tensor:
    """Tensor with arbitrarily dense or sparse dimensions.

    This is a thin Python wrapper around a taco structure stored in C and managed by cffi. An instance should be
    constructed via the `Tensor.from_*` static methods.
    """

    def __init__(self, cffi_tensor):
        self.cffi_tensor = cffi_tensor

    @staticmethod
    def from_lol(
        lol,
        *,
        dimensions: Optional[Tuple[int, ...]] = None,
        format: Union[Format, str, None] = None,
    ) -> Tensor:
        if dimensions is None:
            dimensions = default_lol_dimensions(lol)

        order = len(dimensions)

        if format is None:
            format = Format((Mode.dense,) * order, tuple(range(order)))

        coordinates, values = lol_to_coordinates_and_values(lol)

        return Tensor.from_aos(coordinates, values, dimensions=dimensions, format=format)

    @staticmethod
    def from_dok(
        dictionary: Dict[Tuple[int, ...], float],
        *,
        dimensions: Optional[Tuple[int, ...]] = None,
        format: Union[Format, str, None] = None,
    ) -> Tensor:
        return Tensor.from_aos(
            dictionary.keys(), dictionary.values(), dimensions=dimensions, format=format
        )

    @staticmethod
    def from_aos(
        coordinates: Iterable[Tuple[int, ...]],
        values: Iterable[float],
        *,
        dimensions: Optional[Tuple[int, ...]] = None,
        format: Union[Format, str, None] = None,
    ) -> Tensor:
        # Lengths of modes, dimensions, and elements in coordinates must be equal. Lengths of coordinates and values
        # must be equal
        if dimensions is None:
            dimensions = default_aos_dimensions(coordinates)

        if format is None:
            coordinates = list(coordinates)
            format = default_format_given_nnz(dimensions, len(coordinates))
        elif isinstance(format, str):
            format = parse_format(format).unwrap()

        # Reorder with first level first, etc.
        level_dimensions = tuple(dimensions[i] for i in format.ordering)
        level_coordinates = [
            tuple(coordinate[i] for i in format.ordering) for coordinate in coordinates
        ]

        tree = coordinates_to_tree(level_coordinates, values)

        indexes, vals = tree_to_indices_and_values(tree, format.modes, level_dimensions)

        cffi_modes = tuple(x.c_int for x in format.modes)

        cffi_tensor = taco_structure_to_cffi(
            indexes,
            vals,
            mode_types=cffi_modes,
            dimensions=dimensions,
            mode_ordering=format.ordering,
        )

        return Tensor(cffi_tensor)

    @staticmethod
    def from_soa(
        coordinates: Tuple[Iterable[int], ...],
        values: Iterable[float],
        *,
        dimensions: Optional[Tuple[int, ...]] = None,
        format: Union[Format, str, None] = None,
    ) -> Tensor:
        # Lengths of coordinates, modes, and dimensions must be equal. Lengths of elements of coordinates and values
        # must be equal

        transposed_coordinates = [*zip(*coordinates, strict=True)]

        return Tensor.from_aos(
            transposed_coordinates, values, dimensions=dimensions, format=format
        )

    @staticmethod
    def from_numpy(array, *, format: Union[Format, str, None] = None):
        import numpy

        order = array.ndim
        dimensions = array.shape

        if format is None:
            format = Format((Mode.dense,) * order, tuple(range(order)))

        if numpy.ndim(array) == 0:
            # from_lol does not understand that a scalar numpy array is a scalar
            array = float(array)

        return Tensor.from_lol(array, dimensions=dimensions, format=format)

    @staticmethod
    def from_scipy_sparse(matrix, *, format: Union[Format, str, None] = None):
        import scipy.sparse as scipy_sparse

        if format is None:
            if isinstance(matrix, scipy_sparse.csc_matrix):
                format = Format((Mode.dense, Mode.compressed), (1, 0))
            elif isinstance(matrix, scipy_sparse.csr_matrix):
                format = Format((Mode.dense, Mode.compressed), (0, 1))
            else:
                format = Format((Mode.dense, Mode.compressed), (0, 1))

        soa_matrix = matrix.tocoo()

        return Tensor.from_soa(
            (soa_matrix.row, soa_matrix.col),
            soa_matrix.data,
            dimensions=matrix.shape,
            format=format,
        )

    def to_format(self, format: Union[Format, str]):
        return Tensor.from_dok(self.to_dok(), dimensions=self.dimensions, format=format)

    @property
    def order(self) -> int:
        return self.cffi_tensor.order

    @property
    def dimensions(self) -> Tuple[int, ...]:
        return tuple(self.cffi_tensor.dimensions[0 : self.order])

    @property
    def modes(self) -> Tuple[Mode, ...]:
        return tuple(
            Mode.from_c_int(value) for value in self.cffi_tensor.mode_types[0 : self.order]
        )

    @property
    def mode_ordering(self) -> Tuple[int, ...]:
        return tuple(self.cffi_tensor.mode_ordering[0 : self.order])

    @property
    def taco_indices(self) -> List[List[List[int]]]:
        from .compile import tensor_cdefs

        order = self.order
        dimensions = self.dimensions
        modes = self.modes
        mode_ordering = self.mode_ordering
        cffi_indexes = tensor_cdefs.cast("int32_t***", self.cffi_tensor.indices)

        indices = []
        nnz = 1
        for i_dimension in range(order):
            if modes[i_dimension] == Mode.dense:
                indices.append([])
                nnz *= dimensions[mode_ordering[i_dimension]]
            elif modes[i_dimension] == Mode.compressed:
                pos = list(cffi_indexes[i_dimension][0][0 : nnz + 1])
                crd = list(cffi_indexes[i_dimension][1][0 : pos[-1]])
                indices.append([pos, crd])
                nnz = len(crd)

        return indices

    @property
    def taco_vals(self) -> List[float]:
        from .compile import tensor_cdefs

        order = self.order
        dimensions = self.dimensions
        modes = self.modes
        mode_ordering = self.mode_ordering
        cffi_indexes = tensor_cdefs.cast("int32_t***", self.cffi_tensor.indices)

        nnz = 1
        for i_dimension in range(order):
            if modes[i_dimension] == Mode.dense:
                nnz *= dimensions[mode_ordering[i_dimension]]
            elif modes[i_dimension] == Mode.compressed:
                nnz = cffi_indexes[i_dimension][0][nnz]

        cffi_vals = tensor_cdefs.cast("double*", self.cffi_tensor.vals)
        return list(cffi_vals[0:nnz])

    @property
    def format(self) -> Format:
        return Format(self.modes, self.mode_ordering)

    def items(self) -> Iterator[Tuple[Tuple[int, ...], float]]:
        from .compile import tensor_cdefs

        order = self.order
        modes = self.modes
        dimensions = self.dimensions
        mode_ordering = self.mode_ordering
        cffi_indexes = tensor_cdefs.cast("int32_t***", self.cffi_tensor.indices)
        cffi_values = tensor_cdefs.cast("double*", self.cffi_tensor.vals)
        level_dimensions = [dimensions[i] for i in mode_ordering]

        def recurse(i_level, prefix, position):
            if i_level < order:
                if modes[i_level] == Mode.dense:
                    for index in range(level_dimensions[i_level]):
                        next_position = level_dimensions[i_level] * position + index
                        yield from recurse(i_level + 1, (*prefix, index), next_position)
                elif modes[i_level] == Mode.compressed:
                    start = cffi_indexes[i_level][0][position]
                    end = cffi_indexes[i_level][0][position + 1]

                    for next_position in range(start, end):
                        index = cffi_indexes[i_level][1][next_position]
                        yield from recurse(i_level + 1, (*prefix, index), next_position)
            else:
                coordinate = tuple(prefix[mode_ordering[i]] for i in range(order))
                yield coordinate, cffi_values[position]

        yield from recurse(0, (), 0)

    def to_dok(self, *, explicit_zeros=False) -> Dict[Tuple[int, ...], float]:
        if explicit_zeros:
            return dict(self.items())
        else:
            return {key: value for key, value in self.items() if value != 0.0}

    def to_numpy(self):
        import numpy

        array = numpy.zeros(self.dimensions, dtype=float)

        for index, value in self.items():
            array[index] = value

        return array

    def __add__(self, other) -> Tensor:
        return evaluate_binary_operator(self, other, "+")

    def __radd__(self, other) -> Tensor:
        return evaluate_binary_operator(other, self, "+")

    def __sub__(self, other) -> Tensor:
        return evaluate_binary_operator(self, other, "-")

    def __rsub__(self, other) -> Tensor:
        return evaluate_binary_operator(other, self, "-")

    def __mul__(self, other) -> Tensor:
        return evaluate_binary_operator(self, other, "*")

    def __rmul__(self, other) -> Tensor:
        return evaluate_binary_operator(other, self, "*")

    def __matmul__(self, other):
        return evaluate_matrix_multiplication_operator(self, other)

    def __rmatmul__(self, other):
        return evaluate_matrix_multiplication_operator(other, self)

    def __float__(self):
        from .compile import tensor_cdefs

        if self.order != 0:
            raise ValueError(
                f"Can only convert Tensor of order 0 to float, not order {self.order}"
            )

        cffi_vals = tensor_cdefs.cast("double*", self.cffi_tensor.vals)
        return cffi_vals[0]

    def __getstate__(self):
        return {
            "dimensions": self.dimensions,
            "mode_types": tuple(mode.c_int for mode in self.format.modes),
            "mode_ordering": self.format.ordering,
            "indices": self.taco_indices,
            "vals": self.taco_vals,
        }

    def __setstate__(self, state):
        self.cffi_tensor = taco_structure_to_cffi(
            indices=state["indices"],
            vals=state["vals"],
            mode_types=state["mode_types"],
            dimensions=state["dimensions"],
            mode_ordering=state["mode_ordering"],
        )

    def __eq__(self, other):
        if isinstance(other, Tensor):
            return self.to_dok() == other.to_dok()
        else:
            return NotImplemented

    def __repr__(self):
        return f"Tensor.from_dok({self.to_dok()}, dimensions={self.dimensions}, format={self.format.deparse()!r})"


def lol_to_coordinates_and_values(
    data: Any, keep_zero: bool = False
) -> Tuple[Iterable[Tuple[int, ...]], Iterable[float]]:
    coordinates = []
    values = []

    def recurse(tree: List[Any], indexes: Tuple[int, ...]):
        if isinstance(tree, Real):
            # A leaf was reached
            if keep_zero or tree != 0.0:
                coordinates.append(indexes)
                values.append(tree)
        else:
            for i_element, element in enumerate(tree):
                recurse(element, (*indexes, i_element))

    recurse(data, ())

    return coordinates, values


def coordinates_to_tree(coordinates: Iterable[Tuple[int, ...]], values: Iterable[float]) -> Any:
    tree = None

    def recurse(node: Dict[int, Any], remaining_coordinates: Tuple[int, ...], payload: float):
        key = remaining_coordinates[0]
        if len(remaining_coordinates) == 1:
            node[key] = node.get(key, 0.0) + payload
        else:
            if key not in node:
                node[key] = {}
            recurse(node[key], remaining_coordinates[1:], payload)

    for coordinate, value in zip(coordinates, values, strict=True):
        if len(coordinate) == 0:
            # Coordinates for order-0 tensors must be handled separately from others
            if tree is None:
                tree = 0.0
            tree += value
        else:
            if tree is None:
                tree = {}
            recurse(tree, coordinate, value)

    return tree


def tree_to_indices_and_values(
    tree: Any, modes: Tuple[Mode, ...], dimensions: Tuple[int, ...]
) -> Tuple[List[List[List[int]]], List[float]]:
    order = len(modes)

    # Initialize indexes structure
    indexes = []
    values = []
    for mode, _dimension in zip(modes, dimensions, strict=True):
        if mode == Mode.dense:
            indexes.append([])
        elif mode == Mode.compressed:
            indexes.append([[0], []])

    def recurse(node, i_level):
        if modes[i_level] == Mode.dense:
            iter_next_level = range(dimensions[i_level])
        elif modes[i_level] == Mode.compressed:
            idx = sorted(node.keys())
            indexes[i_level][0].append(indexes[i_level][0][-1] + len(idx))
            indexes[i_level][1].extend(idx)

            iter_next_level = idx
        else:
            raise NotImplementedError()

        if i_level == order - 1:
            # Final level is reached; append values
            for key in iter_next_level:
                values.append(node.get(key, 0.0))
        else:
            # Still descending tree; recurse
            for key in iter_next_level:
                next_tree = node.get(key, {})
                recurse(next_tree, i_level + 1)

    if len(dimensions) == 0:
        if tree is None:
            values = [0.0]
        else:
            values = [tree]
    else:
        if tree is None:
            recurse({}, 0)
        else:
            recurse(tree, 0)

    return indexes, values


def evaluate_binary_operator(
    left: Union[Tensor, Real], right: Union[Tensor, Real], operator: str
) -> Tensor:
    from .compile import evaluate_tensora

    def indexes_string(tensor):
        return ",".join(f"i{i}" for i in range(tensor.order))

    if isinstance(left, Tensor) and isinstance(right, Tensor):
        if left.dimensions != right.dimensions:
            raise ValueError(
                f"Cannot apply operator {operator} between tensor with dimensions {left.dimensions} and "
                f"tensor with dimensions {right.dimensions}"
            )

        if operator == "*":
            # Output has density of least dense tensor
            output_format = "".join(
                "d" if mode1 == Mode.dense and mode2 == Mode.dense else "s"
                for mode1, mode2 in zip(left.format.modes, right.format.modes, strict=True)
            )
        elif operator in ("+", "-"):
            # Output has density of most dense tensor
            output_format = "".join(
                "d" if mode1 == Mode.dense or mode2 == Mode.dense else "s"
                for mode1, mode2 in zip(left.format.modes, right.format.modes, strict=True)
            )
        else:
            raise NotImplementedError()

        indexes = indexes_string(left)
        return evaluate_tensora(
            f"output({indexes}) = left({indexes}) {operator} right({indexes})",
            output_format,
            left=left,
            right=right,
        )

    elif isinstance(left, Tensor) and isinstance(right, Real):
        if operator == "*":
            # Output has density of tensor
            output_format = left.format.deparse()
        elif operator in ("+", "-"):
            # Output is full dense
            output_format = "d" * left.order
        else:
            raise NotImplementedError()

        indexes = indexes_string(left)
        return evaluate_tensora(
            f"output({indexes}) = left({indexes}) {operator} right()",
            output_format,
            left=left,
            right=Tensor.from_lol(float(right)),
        )

    elif isinstance(left, Real) and isinstance(right, Tensor):
        if operator == "*":
            # Output has density of tensor
            output_format = right.format.deparse()
        elif operator in ("+", "-"):
            # Output is full dense
            output_format = "d" * right.order
        else:
            raise NotImplementedError()

        indexes = indexes_string(right)
        return evaluate_tensora(
            f"output({indexes}) = left() {operator} right({indexes})",
            output_format,
            left=Tensor.from_lol(float(left)),
            right=right,
        )

    else:
        return NotImplemented


def evaluate_matrix_multiplication_operator(left: Tensor, right: Tensor):
    from .compile import evaluate_tensora

    if isinstance(left, Tensor) and isinstance(right, Tensor):
        if left.order == 1 and right.order == 1:
            if left.dimensions != right.dimensions:
                raise ValueError(
                    f"Cannot apply operator @ between tensor with dimensions {left.dimensions} and "
                    f"tensor with dimensions {right.dimensions}"
                )

            return evaluate_tensora("output() = left(i) * right(i)", "", left=left, right=right)
        elif left.order == 2 and right.order == 1:
            if left.dimensions[1] != right.dimensions[0]:
                raise ValueError(
                    f"Cannot apply operator @ between tensor with dimensions {left.dimensions} and "
                    f"tensor with dimensions {right.dimensions}"
                )

            # Output format is the uncontracted dimension of the matrix
            output_format = left.format.modes[left.format.ordering[0]].character
            return evaluate_tensora(
                "output(i) = left(i,j) * right(j)", output_format, left=left, right=right
            )
        elif left.order == 1 and right.order == 2:
            if left.dimensions[0] != right.dimensions[0]:
                raise ValueError(
                    f"Cannot apply operator @ between tensor with dimensions {left.dimensions} and "
                    f"tensor with dimensions {right.dimensions}"
                )

            # Output format is the uncontracted dimension of the matrix
            output_format = right.format.modes[right.format.ordering[1]].character
            return evaluate_tensora(
                "output(j) = left(i) * right(i,j)", output_format, left=left, right=right
            )
        elif left.order == 2 and right.order == 2:
            if left.dimensions[1] != right.dimensions[0]:
                raise ValueError(
                    f"Cannot apply operator @ between tensor with dimensions {left.dimensions} and "
                    f"tensor with dimensions {right.dimensions}"
                )

            # Output format are the uncontracted dimensions of the matrices
            left_output_format = left.format.modes[left.format.ordering[0]].character
            right_output_format = right.format.modes[right.format.ordering[1]].character
            output_format = left_output_format + right_output_format
            return evaluate_tensora(
                "output(i,k) = left(i,j) * right(j,k)", output_format, left=left, right=right
            )
        else:
            raise ValueError(
                f"Matrix multiply is only defined between tensors of orders 1 and 2, not orders "
                f"{left.order} and {right.order}"
            )
    else:
        return NotImplemented


def default_lol_dimensions(lol) -> Tuple[int, ...]:
    """Extract dimensions from dense list-of-lists.

    Given nested lists of lists representing a dense tensor in row-major format, discover the dimensions of the tensor
    as implied by the lengths of the lists. The length of the top-level list is the size of the first dimension, the
    length of the first element of that list is the size of the second dimension, and so on until a scalar is
    encountered. For example, `infer_lol_dimensions([[1,2,3],[4,5,6]])` returns `(2,3)`.

    A scalar value is allowed, which translates correctly to dimensions equal to `()`.

    Not all tensors can have their dimensions inferred from a dense LOL. A tensor with dimensions `(0,2)` will never be
    inferred because its dense LOL representation is `[]`, which would actually be inferred as having dimensions `(0,)`.
    """
    dimensions = []
    subdata = lol
    while True:
        if isinstance(subdata, list):
            dimensions.append(len(subdata))
            if len(subdata) > 0:
                subdata = subdata[0]
            else:
                break
        else:
            break

    return tuple(dimensions)


def default_aos_dimensions(coordinates: Iterable[Tuple[int, ...]]) -> Tuple[int, ...]:
    order = None
    maximums = []
    for coordinate in coordinates:
        if order is None:
            order = len(coordinate)
            maximums = list(coordinate)
        else:
            if len(coordinate) != order:
                raise ValueError(
                    f"All coordinates must be the same length; the first coordinate has length"
                    f"{order}, but this coordinate is not that length: {coordinate}"
                )

            for i, (dimension, index) in enumerate(zip(maximums, coordinate, strict=True)):
                if index > dimension:
                    maximums[i] = coordinate[i]

    # i+ 1 is the length of a dimension whose largest index is i
    return tuple(i + 1 for i in maximums)


def default_format_given_nnz(dimensions: Tuple[int, ...], nnz: int) -> Format:
    # The default format is to use dense dimensions as long as the number of nonzeros is larger
    # than the product of those dimensions.
    needed_dense = 0
    required_threshold = 1
    for needed_dense, dimension in enumerate(dimensions):  # noqa: B007; needed_dense is used
        required_threshold *= dimension
        if nnz < required_threshold:
            break

    return Format(
        (Mode.dense,) * needed_dense + (Mode.compressed,) * (len(dimensions) - needed_dense),
        tuple(range(len(dimensions))),
    )


def taco_indexes_from_aos_coordinates(
    coordinates: Iterable[Tuple[int, ...]],
    values: Iterable[float],
    *,
    modes: Tuple[Mode, ...],
    dimensions=Tuple[int, ...],
):  # pragma: no cover
    # This is an experimental alternative to coordinates_to_tree and tree_to_indices_and_values. It is not currently
    # used anywhere.

    class CartesianAppend(Iterable):
        # Does a cartesian product, but assumes
        def __init__(self, base: Iterable[Sequence[Any]], append: Iterable[Any]):
            self.base = base
            self.append = append

        def __iter__(self):
            for prefix, suffix in itertools.product(self.base, self.append):
                yield (*prefix, suffix)

    # Sort coordinates
    sorted_coordinates_and_values = sorted(
        zip(coordinates, values, strict=True), key=lambda x: x[0]
    )

    # Remove duplicates
    deduplicated_coordinates = []
    deduplicated_values = []
    previous_coordinate = None
    for coordinate, value in sorted_coordinates_and_values:
        if coordinate == previous_coordinate:
            deduplicated_values[-1] += value
        else:
            deduplicated_coordinates.append(coordinate)
            deduplicated_values.append(value)

    # To SoA coordinates for easier handling
    soa_coordinates = [*zip(*deduplicated_coordinates, strict=True)]

    # Build taco levels from sorted, deduplicated, SoA coordinates
    levels = []
    previous_prefixes = [()]
    for i_level, (mode, dimension) in enumerate(zip(modes, dimensions, strict=True)):
        if mode == Mode.dense:
            levels.append([dimension])
            previous_prefixes = CartesianAppend(previous_prefixes, range(dimension))
        elif mode == Mode.compressed:
            # There needs to be one entry in pos per previous prefix indicating the position in idx when this prefix
            # is encountered
            previous_prefix_iterator = iter(previous_prefixes)
            current_prefix = None

            pos = []
            idx = []
            unique_coordinates = []
            previous_coordinate = None
            for coordinate in zip(*soa_coordinates[0 : i_level + 1], strict=True):
                if coordinate != previous_coordinate:
                    while coordinate[0:-1] != current_prefix:
                        # The prefix has changed. Mark in pos the position of this prefix. Some prefixes may be
                        # empty (e.g. the prefix came from a dense range); in such a case, we need to emit the
                        # current position multiple times until the current prefix is found.
                        current_prefix = next(previous_prefix_iterator)
                        pos.append(len(idx))

                    # Store coordinate of this level in idx
                    idx.append(coordinate[-1])

                    unique_coordinates.append(coordinate)
                    previous_coordinate = coordinate

            levels.append([pos, idx])
            previous_prefixes = unique_coordinates
        else:
            raise RuntimeError(f"Unknown mode: {mode}")

    return levels



__all__ = [
    "allocate_taco_structure",
    "taco_structure_to_cffi",
    "take_ownership_of_arrays",
    "take_ownership_of_tensor_members",
    "take_ownership_of_tensor",
    "taco_type_header",
    "tensor_cdefs",
]

import platform
from itertools import pairwise
from weakref import WeakKeyDictionary

from cffi import FFI

global_weakkeydict = WeakKeyDictionary()

# order: The number of dimensions of the tensor
# dimensions: The size of each dimension of the tensor; has length `order`
# csize: No idea what this is
# mode_ordering: The dimension that each level refers to; has length `order` and
#   has exactly the numbers 0 to `order` - 1; e.g. if `mode_ordering` is (2, 0, 1), then
#   the first level describes the last dimension, the second level describes the first
#   dimension, etc.
# mode_types: The type (dense, compressed sparse) of each level; importantly, the ith
#   element in `mode_types` describes the ith element in `indices`, not `dimensions`
# indices: A complex data structure storing all the index information of the structure;
#   it has length `order`; each element is one level of the data structure; each level
#   conceptually stores one dimension worth of indexes;
#   *   if a level is dense, then the element in indices is a null pointer or a pointer to
#       a length 0 array or a pointer to a length 1 array, which contains a pointer to a
#       length 1 array, which contains the size of this level's dimension. It does not
#       really matter what goes here because it is never used.
#   *   if a level is compressed (sparse), then the element in indices is a pointer to a
#       length 2 array
#       *   the first element is the `pos` arrays of compressed sparse matrix
#           representation; it has a number of elements equal to the number of indexes
#           in the previous level plus 1; each element `i` in `pos` is the starting
#           position in `idx` for indexes in this dimension associated with the `i`th
#           index of the previous dimension; the last element is the total length of
#           `idx` (that is, the first index not in `idx`)
#       *   the second element is the `idx` array of compressed sparse matrix
#           representation; each element is an index that has a value in this dimension
# vals: The actual values of the sparse matrix; has a length equal to the number of indexes
#   in the last dimension; one has to traverse the `indices` structure to determine the
#   coordinate of each value
# vals_size: Deprecated https://github.com/tensor-compiler/taco/issues/208#issuecomment-476314322

taco_type_header = """
    typedef enum { taco_mode_dense, taco_mode_sparse } taco_mode_t;

    typedef struct {
      int32_t      order;         // tensor order (number of modes)
      int32_t*     dimensions;    // tensor dimensions
      int32_t      csize;         // component size
      int32_t*     mode_ordering; // mode storage ordering
      taco_mode_t* mode_types;    // mode storage types
      int32_t***   indices;       // tensor index data (per mode)
      double*      vals;          // tensor values
      int32_t      vals_size;     // values array size
    } taco_tensor_t;

    void free(void *ptr);
"""

# Define ffi for building the tensors.
tensor_cdefs = FFI()

# This `FFI` is `include`d in each kernel so objects created with its `new` can be passed to the kernels.
tensor_cdefs.cdef(taco_type_header)

# This library only has definitions, in order to `include` it elsewhere, `set_source` must be called with empty `source` first
tensor_cdefs.set_source("_main", "")

if platform.system() == "Windows":
    # On Windows, standard C functions like free are in msvcrt.dll
    tensor_lib = tensor_cdefs.dlopen("msvcrt.dll")
else:
    # On Linux and Mac, standard C functions like free are in the system C library
    tensor_lib = tensor_cdefs.dlopen(None)


def allocate_taco_structure(
    mode_types: tuple[int, ...],
    dimensions: tuple[int, ...],
    mode_ordering: tuple[int, ...],
):
    """Allocate all parts of a taco tensor except growable arrays.

    All int32_t[] tensor.indices[*][*] and double[] tensor.vals are NULL pointers. All other properties are immutable.

    For level i, tensor.indices[i] is
    * if tensor.mode_types[i] is dense: NULL or int32_t*[0]
    * if tensor.mode_types[i] is compressed: int32_t*[2]

    Args:
        mode_types: A list of length order of integer representations of taco_mode_t for each level.
        dimensions: A list of length order of the size of each dimension.
        mode_ordering: A list of length order indicating that the ith mode corresponds to the mode_ordering[i] dimension

    Returns:
        A cffi taco_tensor_t*
    """
    # Validate inputs
    if not (len(mode_types) == len(dimensions) == len(mode_ordering)):
        raise ValueError(
            f"Must all be the same length: mode_types = {mode_types}, dimensions = {dimensions}, "
            f"mode_ordering = {mode_ordering}"
        )

    for mode_type in mode_types:
        if mode_type not in (0, 1):
            raise ValueError(f"mode_types must only contain elements 0 or 1: {mode_types}")

    for dimension in dimensions:
        if dimension < 0:
            # cffi will reject integers too big to fit into an int32_t
            raise ValueError(f"All values in dimensions must be positive: {dimensions}")

    if set(mode_ordering) != set(range(len(mode_types))):
        raise ValueError(
            f"mode_ordering must contain each number in the set {{0, 1, ..., order - 1}} exactly once: "
            f"{mode_ordering}"
        )

    # This structure mimics the taco structure and holds the objects owning the memory of the arrays, ensuring that the
    # pointers stay valid as long as the cffi taco structure has not been garbage collected.
    memory_holder = {}

    cffi_tensor = tensor_cdefs.new("taco_tensor_t*")

    cffi_tensor.order = len(mode_types)

    cffi_dimensions = tensor_cdefs.new("int32_t[]", dimensions)
    memory_holder["dimensions"] = cffi_dimensions
    cffi_tensor.dimensions = cffi_dimensions

    cffi_mode_ordering = tensor_cdefs.new("int32_t[]", mode_ordering)
    memory_holder["mode_ordering"] = cffi_mode_ordering
    cffi_tensor.mode_ordering = cffi_mode_ordering

    cffi_mode_types = tensor_cdefs.new("taco_mode_t[]", mode_types)
    memory_holder["mode_types"] = cffi_mode_types
    cffi_tensor.mode_types = cffi_mode_types

    converted_levels = []
    memory_holder_levels = []
    memory_holder_levels_arrays = []
    for mode in mode_types:
        if mode == tensor_lib.taco_mode_dense:
            converted_arrays = []
        elif mode == tensor_lib.taco_mode_sparse:
            converted_arrays = [tensor_cdefs.NULL, tensor_cdefs.NULL]
        cffi_level = tensor_cdefs.new("int32_t*[]", converted_arrays)
        memory_holder_levels.append(cffi_level)
        memory_holder_levels_arrays.append(converted_arrays)
        converted_levels.append(cffi_level)
    cffi_levels = tensor_cdefs.new("int32_t**[]", converted_levels)
    memory_holder["indices"] = cffi_levels
    memory_holder["*indices"] = memory_holder_levels
    memory_holder["**indices"] = memory_holder_levels_arrays
    cffi_tensor.indices = tensor_cdefs.cast("int32_t***", cffi_levels)

    cffi_tensor.vals = tensor_cdefs.NULL

    cffi_tensor.vals_size = 0

    global_weakkeydict[cffi_tensor] = memory_holder

    return cffi_tensor


def taco_structure_to_cffi(
    indices: list[list[list[int]]],
    vals: list[float],
    *,
    mode_types: tuple[int, ...],
    dimensions: tuple[int, ...],
    mode_ordering: tuple[int, ...],
):
    """Build a cffi taco tensor from Python data.

    This takes Python data with a one-to-one mapping to taco tensor attributes and builds a cffi taco tensor from it.

    Args:
        indices: The list of length order containing the index data required by the taco structure.
        vals: The explicit values required by the taco structure.
        mode_types: A list of length order of integer representations of taco_mode_t for each level.
        dimensions: A list of length order of the size of each dimension.
        mode_ordering: A list of length order indicating that the ith mode corresponds to the mode_ordering[i] dimension

    Returns:
        A cffi taco_tensor_t*. This object owns all the data under it. The associated arrays will be freed when this
        object is garbage collected.
    """
    cffi_tensor = allocate_taco_structure(mode_types, dimensions, mode_ordering)

    # Validate inputs
    if len(indices) != len(mode_types):
        raise ValueError(
            f"Length of indices ({len(indices)}) must be equal to the length of mode_types, dimensions, "
            f"and mode_ordering ({len(mode_types)})"
        )

    nnz = 1
    for i_level in range(cffi_tensor.order):
        if mode_types[i_level] == 0:
            if len(indices[i_level]) != 0:
                raise ValueError(
                    f"Level {i_level} is a dense mode and therefore expects indices[{i_level}] to be "
                    f"empty: {indices[i_level]}"
                )
            nnz *= dimensions[mode_ordering[i_level]]
        elif mode_types[i_level] == 1:
            if len(indices[i_level]) != 2:
                raise ValueError(
                    f"Level {i_level} is a compressed mode and therefore expects indices[{i_level}] to be "
                    f"length 2 not length {len(indices[i_level])}: {indices[i_level]}"
                )
            pos = indices[i_level][0]
            crd = indices[i_level][1]
            if len(pos) != nnz + 1:
                raise ValueError(
                    f"The pos array of level {i_level} (indices[{i_level}][0]) must have length {nnz}, "
                    f"the number of explicit indexes so far, not length {len(pos)}: {pos}"
                )
            if pos[0] != 0:
                raise ValueError(
                    f"The first element of the pos array of level {i_level} must be 0: {pos}"
                )
            if not weakly_increasing(pos):
                raise ValueError(
                    f"The pos array of level {i_level} (indices[{i_level}][0]) must be weakly "
                    f"monotonically increasing: {pos}"
                )
            if len(crd) != pos[-1]:
                raise ValueError(
                    f"The crd array of level {i_level} (indices[{i_level}][1]) must have length "
                    f"{pos[-1]}, the last element of this level's pos array, not length {len(crd)}: {crd}"
                )
            if not all(0 <= x < dimensions[mode_ordering[i_level]] for x in crd):
                raise ValueError(
                    f"All values in the crd array of level {i_level} (indices[{i_level}][1]) must be "
                    f"nonnegative and less than the size of this dimension: {crd}"
                )
            nnz = len(crd)

    if len(vals) != nnz:
        raise ValueError(
            f"Length of vals must be equal to the number of indexes implicitly defined by indices {nnz} "
            f"not {len(vals)}: {vals}"
        )

    # Get the partial constructed memory holder stored by allocate_taco_structure
    memory_holder = global_weakkeydict[cffi_tensor]

    cffi_indices = tensor_cdefs.cast("int32_t***", cffi_tensor.indices)
    for i_level, (mode, level) in enumerate(zip(mode_types, indices, strict=True)):
        if mode == tensor_lib.taco_mode_dense:
            pass
        elif mode == tensor_lib.taco_mode_sparse:
            for i_array, array in enumerate(level):
                cffi_array = tensor_cdefs.new("int32_t[]", array)
                memory_holder["**indices"][i_level][i_array] = cffi_array
                cffi_indices[i_level][i_array] = cffi_array

    cffi_vals = tensor_cdefs.new("double[]", vals)
    memory_holder["vals"] = cffi_vals
    cffi_tensor.vals = tensor_cdefs.cast("double*", cffi_vals)

    cffi_tensor.vals_size = len(vals)

    global_weakkeydict[cffi_tensor] = memory_holder

    return cffi_tensor


def take_ownership_of_arrays(cffi_tensor) -> None:
    """Take ownership of arrays allocated in taco kernel.

    The `evaluate` function created by taco `malloc`s several arrays on the output tensor, namely the `pos` and `crd`
    arrays of compressed levels and the `vals` array. This function looks up the tensor in `global_weakkeydict` and
    attaches the allocated arrays so that they get deallocated when the tensor is garbage collected. This function must
    be called on the output tensor of any evaluated taco kernel.

    Args:
        cffi_tensor: A cffi taco_tensor_t or taco_tensor_t*.
    """

    memory_holder = global_weakkeydict[cffi_tensor]

    order = cffi_tensor.order

    modes = cffi_tensor.mode_types[0:order]

    cffi_levels = tensor_cdefs.cast("int32_t***", cffi_tensor.indices)
    for i_dimension, mode in enumerate(modes):
        if mode == tensor_lib.taco_mode_dense:
            pass
        if mode == tensor_lib.taco_mode_sparse:
            memory_holder["**indices"][i_dimension][0] = tensor_cdefs.gc(
                cffi_levels[i_dimension][0], tensor_lib.free
            )
            memory_holder["**indices"][i_dimension][1] = tensor_cdefs.gc(
                cffi_levels[i_dimension][1], tensor_lib.free
            )

    memory_holder["vals"] = tensor_cdefs.gc(cffi_tensor.vals, tensor_lib.free)


def take_ownership_of_tensor_members(cffi_tensor) -> None:
    """Take ownership of taco tensor whose members were allocated elsewhere.

    This is not needed by pure Tensora. But some low-level programs might
    allocate members of a tensor and return it to Python. If the taco_tensor_t
    will be owned by Python, then this function can be used to take ownership of
    all the memory of all members of the taco_tensor_t so that it is freed when
    the cffi object is freed.

    Args:
        cffi_tensor: A cffi taco_tensor_t or taco_tensor_t*.
    """
    # This mayor may not have been called by take_ownership_of_tensor first, so
    # an entry in global_weakkeydict may or may not already exist.
    memory_holder = global_weakkeydict.get(cffi_tensor, {})

    # First, take ownership of everything that is owned after taco_structure_to_cffi
    order = cffi_tensor.order

    memory_holder["dimensions"] = tensor_cdefs.gc(cffi_tensor.dimensions, tensor_lib.free)

    memory_holder["mode_ordering"] = tensor_cdefs.gc(cffi_tensor.mode_ordering, tensor_lib.free)

    memory_holder["mode_types"] = tensor_cdefs.gc(cffi_tensor.mode_types, tensor_lib.free)

    memory_holder_levels = []
    memory_holder_levels_arrays = []
    for i_dimension, mode in enumerate(cffi_tensor.mode_types[0:order]):
        memory_holder_levels.append(
            tensor_cdefs.gc(cffi_tensor.indices[i_dimension], tensor_lib.free)
        )
        if mode == tensor_lib.taco_mode_dense:
            memory_holder_levels_arrays.append([])
        elif mode == tensor_lib.taco_mode_sparse:
            # It does not matter what the values are here. They will be overwritten in take_ownership_of_arrays.
            memory_holder_levels_arrays.append([tensor_cdefs.NULL, tensor_cdefs.NULL])
    memory_holder["indices"] = tensor_cdefs.gc(cffi_tensor.indices, tensor_lib.free)
    memory_holder["*indices"] = memory_holder_levels
    memory_holder["**indices"] = memory_holder_levels_arrays

    global_weakkeydict[cffi_tensor] = memory_holder

    # Second, defer to take_ownership_of_arrays to take ownership of everything else
    # This function assumes that the memory holder has already been assigned to
    # the weak key dictionary.
    take_ownership_of_arrays(cffi_tensor)


def take_ownership_of_tensor(cffi_tensor) -> None:
    """Take ownership of taco tensor that was allocated elsewhere entirely.

    This is not needed by pure Tensora. But some low-level programs might
    allocate a taco tensor and return it to Python. If the taco_tensor_t will be
    owned by Python, then this function can be used to take ownership of all the
    memory so that it is freed when the cffi object is freed.

    Args:
        cffi_tensor: A cffi taco_tensor_t*.
    """
    global_weakkeydict[cffi_tensor] = {"tensor": tensor_cdefs.gc(cffi_tensor, tensor_lib.free)}
    take_ownership_of_tensor_members(cffi_tensor)


def weakly_increasing(list: list[int]):
    return all(x <= y for x, y in pairwise(list))



__all__ = ["Definition", "TensorDimension"]

from dataclasses import dataclass

from ..format import Format
from .identifiable_expression.ast import Tensor


@dataclass(frozen=True, slots=True)
class TensorDimension:
    name: str
    dimension: int


@dataclass(frozen=True, slots=True)
class Definition:
    output_variable: Tensor
    formats: dict[str, Format]
    indexes: dict[str, TensorDimension]



from __future__ import annotations

__all__ = ["extract_context"]

from dataclasses import dataclass, field
from functools import singledispatch

from ...format import Mode
from . import ast
from ._tensor_layer import TensorLayer


@dataclass(frozen=True, slots=True, kw_only=True)
class Context:
    is_sparse: bool
    sparse_leaves: list[TensorLayer] = field(default_factory=list)
    dense_leaves: list[TensorLayer] = field(default_factory=list)
    indexes: frozenset[str] = frozenset()
    has_output: bool = False
    has_assemble: bool = False

    def add(self, other: Context) -> Context:
        return Context(
            is_sparse=self.is_sparse and other.is_sparse,
            sparse_leaves=self.sparse_leaves + other.sparse_leaves,
            dense_leaves=self.dense_leaves + other.dense_leaves,
            indexes=self.indexes | other.indexes,
            has_output=self.has_output or other.has_output,
            has_assemble=self.has_assemble or other.has_assemble,
        )

    def multiply(self, other: Context) -> Context:
        return Context(
            is_sparse=self.is_sparse or other.is_sparse,
            sparse_leaves=self.sparse_leaves + other.sparse_leaves,
            dense_leaves=self.dense_leaves + other.dense_leaves,
            indexes=self.indexes | other.indexes,
            has_output=self.has_output or other.has_output,
            has_assemble=self.has_assemble or other.has_assemble,
        )


@singledispatch
def extract_context(self: ast.Expression, index: str) -> Context:
    raise NotImplementedError(f"extract_context not implemented for {type(self)}: {self}")


@extract_context.register(ast.Literal)
def extract_context_literal(self: ast.Literal, index: str) -> Context:
    if self == ast.Integer(0) or self == ast.Float(0.0):
        return Context(is_sparse=True)
    else:
        return Context(is_sparse=False)


@extract_context.register(ast.Tensor)
def extract_context_tensor(self: ast.Tensor, index: str) -> Context:
    try:
        maybe_layer = self.indexes.index(index)
    except ValueError:
        return Context(is_sparse=False)

    if self.modes[maybe_layer] == Mode.dense:
        return Context(is_sparse=False, dense_leaves=[TensorLayer(self, maybe_layer)])
    else:
        return Context(is_sparse=True, sparse_leaves=[TensorLayer(self, maybe_layer)])


@extract_context.register(ast.Add)
def extract_context_add(self: ast.Add, index: str) -> Context:
    left = extract_context(self.left, index)
    right = extract_context(self.right, index)
    return left.add(right)


@extract_context.register(ast.Multiply)
def extract_context_multiply(self: ast.Multiply, index: str) -> Context:
    left = extract_context(self.left, index)
    right = extract_context(self.right, index)
    return left.multiply(right)



__all__ = ["MutatingAssignmentError", "InconsistentDimensionsError", "NameConflictError"]

from dataclasses import dataclass

from .ast import Assignment, Tensor


@dataclass(frozen=True, slots=True)
class MutatingAssignmentError(Exception):
    assignment: Assignment

    def __str__(self):
        return (
            f"Expected assignment target to never appear on the right hand side, "
            f"but found {self.assignment.target.name} on both sides of {self.assignment}"
        )


@dataclass(frozen=True, slots=True)
class InconsistentDimensionsError(Exception):
    assignment: Assignment
    first: Tensor
    second: Tensor

    def __str__(self):
        return (
            f"Expected each tensor in an assignment to be referenced with the same number of "
            f"indexes, but found parameter {self.first.name} referenced as {self.first} and then "
            f"as {self.second} in {self.assignment}"
        )


@dataclass(frozen=True, slots=True)
class NameConflictError(Exception):
    name: str
    assignment: Assignment

    def __str__(self):
        return (
            f"Expected no tensor and index to have the same name, but found {self.name} as both a "
            f"tensor and an index in {self.assignment}"
        )



__all__ = ["ir_to_llvm"]

from functools import singledispatch

import llvmlite.ir as llvm

from ..compile import target
from ..ir.ast import (
    Add,
    And,
    ArrayAllocate,
    ArrayIndex,
    ArrayReallocate,
    Assignable,
    Assignment,
    AttributeAccess,
    Block,
    BooleanLiteral,
    BooleanToInteger,
    Branch,
    Declaration,
    DeclarationAssignment,
    Equal,
    Expression,
    FloatLiteral,
    FunctionDefinition,
    GreaterThan,
    GreaterThanOrEqual,
    IntegerLiteral,
    LessThan,
    LessThanOrEqual,
    Loop,
    Max,
    Min,
    Module,
    Multiply,
    NotEqual,
    Or,
    Return,
    Statement,
    Subtract,
    Variable,
)
from ._hoist_declarations import hoist_declarations
from ._type_to_llvm import (
    attribute_indexes,
    llvm_boolean_type,
    llvm_float_type,
    llvm_integer_type,
    llvm_size_type,
    type_to_llvm,
)

# Create a target machine from the target so that we can get the word sizes
target_machine = target.create_target_machine()


@singledispatch
def get_element_pointer(
    self: Assignable, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    raise NotImplementedError(f"get_element_pointer not implemented for {type(self)}: {self}")


@get_element_pointer.register(Variable)
def get_element_pointer_variable(
    self: Variable, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    return locals[self.name]


@get_element_pointer.register(AttributeAccess)
def get_element_pointer_attribute_access(
    self: AttributeAccess, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    stem = ir_to_llvm_expression(self.target, builder, locals)
    return builder.gep(
        stem,
        [
            llvm.Constant(llvm_size_type, 0),
            # All structs are indexed with i32 in LLVM
            llvm.Constant(llvm.IntType(32), attribute_indexes[self.attribute]),
        ],
    )


@get_element_pointer.register(ArrayIndex)
def get_element_pointer_array_index(
    self: ArrayIndex, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    target = ir_to_llvm_expression(self.target, builder, locals)
    index = ir_to_llvm_expression(self.index, builder, locals)
    return builder.gep(target, [index])


@singledispatch
def ir_to_llvm_expression(
    self: Expression, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    raise NotImplementedError(f"ir_to_llvm_expression not implemented for {type(self)}: {self}")


@ir_to_llvm_expression.register(Variable)
def ir_to_llvm_variable(
    self: Variable, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    return builder.load(locals[self.name])


@ir_to_llvm_expression.register(AttributeAccess)
def ir_to_llvm_attribute_access(
    self: AttributeAccess, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    stem = ir_to_llvm_expression(self.target, builder, locals)
    return builder.load(
        builder.gep(
            stem,
            [
                llvm.Constant(llvm_size_type, 0),
                # All structs are indexed with i32 in LLVM
                llvm.Constant(llvm.IntType(32), attribute_indexes[self.attribute]),
            ],
        )
    )


@ir_to_llvm_expression.register(ArrayIndex)
def ir_to_llvm_array_index(
    self: ArrayIndex, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    target = ir_to_llvm_expression(self.target, builder, locals)
    index = ir_to_llvm_expression(self.index, builder, locals)
    return builder.load(builder.gep(target, [index]))


@ir_to_llvm_expression.register(IntegerLiteral)
def ir_to_llvm_integer_literal(
    self: IntegerLiteral, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    return llvm.Constant(llvm_integer_type, self.value)


@ir_to_llvm_expression.register(FloatLiteral)
def ir_to_llvm_float_literal(
    self: FloatLiteral, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    return llvm.Constant(llvm_float_type, self.value)


@ir_to_llvm_expression.register(BooleanLiteral)
def ir_to_llvm_boolean_literal(
    self: BooleanLiteral, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    return llvm.Constant(llvm_boolean_type, self.value)


@ir_to_llvm_expression.register(Add)
def ir_to_llvm_add(
    self: Add, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    left = ir_to_llvm_expression(self.left, builder, locals)
    right = ir_to_llvm_expression(self.right, builder, locals)
    match left.type, right.type:
        case (llvm.IntType(), llvm.IntType()):
            return builder.add(left, right)
        case (llvm.IntType(), llvm.DoubleType()):
            left = builder.sitofp(left, llvm_float_type)
            return builder.fadd(left, right)
        case (llvm.DoubleType(), llvm.IntType()):
            right = builder.sitofp(right, llvm_float_type)
            return builder.fadd(left, right)
        case (llvm.DoubleType(), llvm.DoubleType()):
            return builder.fadd(left, right)
        case (llvm.PointerType(), llvm.IntType()):
            return builder.gep(left, [right])
        case _:
            raise TypeError(f"Cannot add {left.type} and {right.type}")


@ir_to_llvm_expression.register(Subtract)
def ir_to_llvm_subtract(
    self: Subtract, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    left = ir_to_llvm_expression(self.left, builder, locals)
    right = ir_to_llvm_expression(self.right, builder, locals)
    match left.type, right.type:
        case (llvm.IntType(), llvm.IntType()):
            return builder.sub(left, right)
        case (llvm.IntType(), llvm.DoubleType()):
            left = builder.sitofp(left, llvm_float_type)
            return builder.fsub(left, right)
        case (llvm.DoubleType(), llvm.IntType()):
            right = builder.sitofp(right, llvm_float_type)
            return builder.fsub(left, right)
        case (llvm.DoubleType(), llvm.DoubleType()):
            return builder.fsub(left, right)
        case _:
            raise TypeError(f"Cannot add {left.type} and {right.type}")


@ir_to_llvm_expression.register(Multiply)
def ir_to_llvm_multiply(
    self: Multiply, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    left = ir_to_llvm_expression(self.left, builder, locals)
    right = ir_to_llvm_expression(self.right, builder, locals)
    match left.type, right.type:
        case (llvm.IntType(), llvm.IntType()):
            return builder.mul(left, right)
        case (llvm.IntType(), llvm.DoubleType()):
            left = builder.sitofp(left, llvm_float_type)
            return builder.fmul(left, right)
        case (llvm.DoubleType(), llvm.IntType()):
            right = builder.sitofp(right, llvm_float_type)
            return builder.fmul(left, right)
        case (llvm.DoubleType(), llvm.DoubleType()):
            return builder.fmul(left, right)
        case _:
            raise TypeError(f"Cannot add {left.type} and {right.type}")


@ir_to_llvm_expression.register(Equal)
def ir_to_llvm_equal(
    self: Equal, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    return builder.icmp_signed(
        "==",
        ir_to_llvm_expression(self.left, builder, locals),
        ir_to_llvm_expression(self.right, builder, locals),
    )


@ir_to_llvm_expression.register(NotEqual)
def ir_to_llvm_not_equal(
    self: NotEqual, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    return builder.icmp_signed(
        "!=",
        ir_to_llvm_expression(self.left, builder, locals),
        ir_to_llvm_expression(self.right, builder, locals),
    )


@ir_to_llvm_expression.register(GreaterThan)
def ir_to_llvm_greater_than(
    self: GreaterThan, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    return builder.icmp_signed(
        ">",
        ir_to_llvm_expression(self.left, builder, locals),
        ir_to_llvm_expression(self.right, builder, locals),
    )


@ir_to_llvm_expression.register(LessThan)
def ir_to_llvm_less_than(
    self: LessThan, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    return builder.icmp_signed(
        "<",
        ir_to_llvm_expression(self.left, builder, locals),
        ir_to_llvm_expression(self.right, builder, locals),
    )


@ir_to_llvm_expression.register(GreaterThanOrEqual)
def ir_to_llvm_greater_than_or_equal(
    self: GreaterThanOrEqual, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    return builder.icmp_signed(
        ">=",
        ir_to_llvm_expression(self.left, builder, locals),
        ir_to_llvm_expression(self.right, builder, locals),
    )


@ir_to_llvm_expression.register(LessThanOrEqual)
def ir_to_llvm_less_than_or_equal(
    self: LessThanOrEqual, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    return builder.icmp_signed(
        "<=",
        ir_to_llvm_expression(self.left, builder, locals),
        ir_to_llvm_expression(self.right, builder, locals),
    )


@ir_to_llvm_expression.register(And)
def ir_to_llvm_and(
    self: And, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    left_block = builder.block
    right_block = builder.append_basic_block()
    end_block = builder.append_basic_block()

    left = ir_to_llvm_expression(self.left, builder, locals)

    builder.cbranch(left, right_block, end_block)

    builder.position_at_end(right_block)
    right = ir_to_llvm_expression(self.right, builder, locals)
    builder.branch(end_block)

    builder.position_at_end(end_block)
    phi = builder.phi(llvm_boolean_type)
    phi.add_incoming(llvm.Constant(llvm_boolean_type, 0), left_block)
    phi.add_incoming(right, right_block)

    return phi


@ir_to_llvm_expression.register(Or)
def ir_to_llvm_or(self: Or, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]) -> llvm.Value:
    left_block = builder.block
    right_block = builder.append_basic_block()
    end_block = builder.append_basic_block()

    left = ir_to_llvm_expression(self.left, builder, locals)

    builder.cbranch(left, end_block, right_block)

    builder.position_at_end(right_block)
    right = ir_to_llvm_expression(self.right, builder, locals)
    builder.branch(end_block)

    builder.position_at_end(end_block)
    phi = builder.phi(llvm_boolean_type)
    phi.add_incoming(llvm.Constant(llvm_boolean_type, 1), left_block)
    phi.add_incoming(right, right_block)

    return phi


@ir_to_llvm_expression.register(Max)
def ir_to_llvm_max(
    self: Max, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    left = ir_to_llvm_expression(self.left, builder, locals)
    right = ir_to_llvm_expression(self.right, builder, locals)
    condition = builder.icmp_signed(">", left, right)
    return builder.select(condition, left, right)


@ir_to_llvm_expression.register(Min)
def ir_to_llvm_min(
    self: Min, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    left = ir_to_llvm_expression(self.left, builder, locals)
    right = ir_to_llvm_expression(self.right, builder, locals)
    condition = builder.icmp_signed("<", left, right)
    return builder.select(condition, left, right)


@ir_to_llvm_expression.register(BooleanToInteger)
def ir_to_llvm_boolean_to_integer(
    self: BooleanToInteger, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    expression = ir_to_llvm_expression(self.expression, builder, locals)
    return builder.zext(expression, llvm_integer_type)


@ir_to_llvm_expression.register(ArrayAllocate)
def ir_to_llvm_array_allocate(
    self: ArrayAllocate, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    element_size = llvm.Constant(
        llvm_integer_type, type_to_llvm(self.element_type).get_abi_size(target_machine.target_data)
    )
    n_elements = ir_to_llvm_expression(self.n_elements, builder, locals)
    memory_size = builder.zext(builder.mul(element_size, n_elements), llvm_size_type)
    memory_pointer = builder.call(locals["malloc"], [memory_size])
    array_pointer = builder.bitcast(memory_pointer, type_to_llvm(self.element_type).as_pointer())
    return array_pointer


@ir_to_llvm_expression.register(ArrayReallocate)
def ir_to_llvm_array_reallocate(
    self: ArrayReallocate, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    old_array_pointer = ir_to_llvm_expression(self.old, builder, locals)
    old_memory_pointer = builder.bitcast(old_array_pointer, llvm.IntType(8).as_pointer())
    element_size = llvm.Constant(
        llvm_integer_type, type_to_llvm(self.element_type).get_abi_size(target_machine.target_data)
    )
    n_elements = ir_to_llvm_expression(self.n_elements, builder, locals)
    memory_size = builder.zext(builder.mul(element_size, n_elements), llvm_size_type)
    memory_pointer = builder.call(locals["realloc"], [old_memory_pointer, memory_size])
    array_pointer = builder.bitcast(memory_pointer, type_to_llvm(self.element_type).as_pointer())
    return array_pointer


def ir_to_llvm_declaration(
    self: Declaration, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> llvm.Value:
    # All declarations are hoisted in LLVM, so this merely fetches the previously declared variable
    return locals[self.name.name]


@singledispatch
def ir_to_llvm_statement(
    self: Statement, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> None:
    raise NotImplementedError(f"ir_to_llvm_statement not implemented for {type(self)}: {self}")


@ir_to_llvm_statement.register(Expression)
def convert_expression_to_statement(
    self: Expression, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> None:
    # An expression can also be a statement; run it here and throw away the result
    _ = ir_to_llvm_expression(self)


@ir_to_llvm_statement.register(Declaration)
def convert_declaration_to_statement(
    self: Declaration, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> None:
    # A declaration can also be a statement; declare the variable and throw away the result
    _ = ir_to_llvm_declaration(self, builder, locals)


@ir_to_llvm_statement.register(Assignment)
def ir_to_llvm_assignment(
    self: Assignment, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> None:
    value = ir_to_llvm_expression(self.value, builder, locals)
    target = get_element_pointer(self.target, builder, locals)
    match value.type, target.type.pointee:
        case (llvm.IntType(), llvm.DoubleType()):
            value = builder.sitofp(value, llvm_float_type)
        case _:
            pass
    builder.store(value, target)


@ir_to_llvm_statement.register(DeclarationAssignment)
def ir_to_llvm_declaration_assignment(
    self: DeclarationAssignment, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> None:
    value = ir_to_llvm_expression(self.value, builder, locals)
    target = ir_to_llvm_declaration(self.target, builder, locals)
    match value.type, target.type:
        case (llvm.IntType(), llvm.DoubleType()):
            value = builder.sitofp(value, llvm_float_type)
        case _:
            pass
    builder.store(value, target)


@ir_to_llvm_statement.register(Block)
def ir_to_llvm_block(self: Block, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]) -> None:
    if self.comment is not None:
        # Add comment if it there is one
        builder.comment(self.comment)

    for statement in self.statements:
        ir_to_llvm_statement(statement, builder, locals)


@ir_to_llvm_statement.register(Branch)
def ir_to_llvm_branch(
    self: Branch, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> None:
    condition = ir_to_llvm_expression(self.condition, builder, locals)
    with builder.if_else(condition) as (then, otherwise):
        with then:
            ir_to_llvm_statement(self.if_true, builder, locals)
        with otherwise:
            ir_to_llvm_statement(self.if_false, builder, locals)


@ir_to_llvm_statement.register(Loop)
def ir_to_llvm_loop(self: Loop, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]) -> None:
    condition_block = builder.append_basic_block()
    body_block = builder.append_basic_block()
    end_block = builder.append_basic_block()

    builder.branch(condition_block)

    builder.position_at_end(condition_block)
    condition = ir_to_llvm_expression(self.condition, builder, locals)
    builder.cbranch(condition, body_block, end_block)

    builder.position_at_end(body_block)
    ir_to_llvm_statement(self.body, builder, locals)
    builder.branch(condition_block)

    builder.position_at_end(end_block)


@ir_to_llvm_statement.register(Return)
def ir_to_llvm_return(
    self: Return, builder: llvm.IRBuilder, locals: dict[str, llvm.Value]
) -> None:
    builder.ret(ir_to_llvm_expression(self.value, builder, locals))


def ir_to_llvm_function_definition(
    self: FunctionDefinition, module: llvm.Module, functions: dict[str, llvm.Function]
) -> None:
    return_type = type_to_llvm(self.return_type)
    argument_types = [type_to_llvm(parameter.type) for parameter in self.parameters]
    function_type = llvm.FunctionType(return_type, argument_types)
    function = llvm.Function(module, function_type, name=self.name.name)

    body = llvm.IRBuilder(function.append_basic_block())
    locals = {}

    # Assign each function parameter to a local variable
    for parameter, llvm_parameter in zip(self.parameters, function.args, strict=False):
        variable = body.alloca(type_to_llvm(parameter.type), name=parameter.name.name)
        body.store(llvm_parameter, variable)
        locals[parameter.name.name] = variable

    # Hoist all declarations in the function body
    for name, type in hoist_declarations(self).items():
        variable = body.alloca(type_to_llvm(type), name=name)
        locals[name] = variable

    ir_to_llvm_statement(self.body, body, locals | functions)


def ir_to_llvm(self: Module) -> llvm.Module:
    module = llvm.Module()

    # Declare malloc
    malloc_signature = llvm.FunctionType(llvm.IntType(8).as_pointer(), [llvm_size_type])
    malloc = llvm.Function(module, malloc_signature, name="malloc")

    # Declare realloc
    realloc_signature = llvm.FunctionType(
        llvm.IntType(8).as_pointer(), [llvm.IntType(8).as_pointer(), llvm_size_type]
    )
    realloc = llvm.Function(module, realloc_signature, name="realloc")

    # This is not compatible with defined functions ever being called
    functions = {"malloc": malloc, "realloc": realloc}

    for function in self.definitions:
        ir_to_llvm_function_definition(function, module, functions)

    return module



from hypothesis import assume, given
from hypothesis import strategies as st
from typer.testing import CliRunner

from tensora.cli import app

runner = CliRunner()


@given(command=st.lists(st.text()))
def test_cli_cannot_crash(command):
    # Arguments to a CLI cannot contain null bytes.
    assume(not any("\0" in string for string in command))

    _ = runner.invoke(app, command, catch_exceptions=False)



from ._best_algorithm import best_algorithm
from ._desugar_expression import desugar_assignment
from ._exceptions import DiagonalAccessError, NoKernelFoundError
from ._index_dimensions import index_dimensions
from ._to_identifiable import to_identifiable



__all__ = ["compile_evaluate"]

import re
import tempfile
import threading
from typing import Any

from cffi import FFI

from ._cffi_ownership import taco_type_header, tensor_cdefs

lock = threading.Lock()

taco_define_header = """
    #ifndef TACO_C_HEADERS
    #define TACO_C_HEADERS
    #define TACO_MIN(_a,_b) ((_a) < (_b) ? (_a) : (_b))
    #define TACO_MAX(_a,_b) ((_a) > (_b) ? (_a) : (_b))
    #endif
"""


def compile_evaluate(source: str) -> Any:
    """Compile evaluate kernel in C code using CFFI.

    Args:
        source: C code containing one evaluate function

    Returns:
        The compiled FFILibrary which has a single method `evaluate` which
        expects cffi pointers to taco_tensor_t instances.
    """
    # Extract signature
    # This needs to be provided alone to cdef
    signature_match = re.search(r"int(32_t)? evaluate\(([^)]*)\)", source)
    signature = signature_match.group(0)

    # Use cffi to compile the kernels
    ffibuilder = FFI()
    ffibuilder.include(tensor_cdefs)
    ffibuilder.cdef(signature + ";")
    ffibuilder.set_source(
        "taco_kernel",
        taco_define_header + taco_type_header + source,
        extra_compile_args=["-Wno-unused-variable", "-Wno-unknown-pragmas"],
    )

    with tempfile.TemporaryDirectory() as temp_dir:
        # Lock because FFI.compile is not thread safe: https://foss.heptapod.net/pypy/cffi/-/issues/490
        with lock:
            # Create shared object in temporary directory
            lib_path = ffibuilder.compile(tmpdir=temp_dir)

        # Load the shared object
        lib = ffibuilder.dlopen(lib_path)

    # Return the entire library rather than just the function because it appears that the memory containing the compiled
    # code is freed as soon as the library goes out of scope: https://stackoverflow.com/q/55323592/1485877
    return lib



from textwrap import dedent

import pytest

from tensora.codegen import ir_to_c, ir_to_c_function_definition, ir_to_c_statement
from tensora.ir.ast import *
from tensora.ir.types import *


def clean(string: str) -> str:
    return dedent(string).strip()


single_lines = [
    # Assignables and literals
    (Variable("x"), "x"),
    (ArrayIndex(Variable("x"), Variable("i")), "x[i]"),
    (ArrayIndex(ArrayIndex(Variable("x"), Variable("i")), Variable("j")), "x[i][j]"),
    (AttributeAccess(Variable("x"), "modes"), "x->modes"),
    (IntegerLiteral(2), "2"),
    (FloatLiteral(1.0), "1.0"),
    (BooleanLiteral(False), "0"),
    (BooleanLiteral(True), "1"),
    # Add
    (Add(Variable("x"), Variable("y")), "x + y"),
    (Add(Add(Variable("x"), Variable("y")), Variable("z")), "x + y + z"),
    (Add(Variable("x"), Add(Variable("y"), Variable("z"))), "x + y + z"),
    (Add(Subtract(Variable("x"), Variable("y")), Variable("z")), "x - y + z"),
    (Add(Variable("x"), Subtract(Variable("y"), Variable("z"))), "x + y - z"),
    (Add(Multiply(Variable("x"), Variable("y")), Variable("z")), "x * y + z"),
    (Add(Variable("x"), Multiply(Variable("y"), Variable("z"))), "x + y * z"),
    # Subtract
    (Subtract(Variable("x"), Variable("y")), "x - y"),
    (Subtract(Add(Variable("x"), Variable("y")), Variable("z")), "x + y - z"),
    (Subtract(Variable("x"), Add(Variable("y"), Variable("z"))), "x - (y + z)"),
    (Subtract(Subtract(Variable("x"), Variable("y")), Variable("z")), "x - y - z"),
    (Subtract(Variable("x"), Subtract(Variable("y"), Variable("z"))), "x - (y - z)"),
    (Subtract(Multiply(Variable("x"), Variable("y")), Variable("z")), "x * y - z"),
    (Subtract(Variable("x"), Multiply(Variable("y"), Variable("z"))), "x - y * z"),
    # Multiply
    (Multiply(Variable("x"), Variable("y")), "x * y"),
    (Multiply(Multiply(Variable("x"), Variable("y")), Variable("z")), "x * y * z"),
    (Multiply(Variable("x"), Multiply(Variable("y"), Variable("z"))), "x * y * z"),
    (Multiply(Add(Variable("x"), Variable("y")), Variable("z")), "(x + y) * z"),
    (Multiply(Variable("x"), Add(Variable("y"), Variable("z"))), "x * (y + z)"),
    (Multiply(Subtract(Variable("x"), Variable("y")), Variable("z")), "(x - y) * z"),
    (Multiply(Variable("x"), Subtract(Variable("y"), Variable("z"))), "x * (y - z)"),
    # Comparison
    (Equal(Variable("x"), Variable("y")), "x == y"),
    (NotEqual(Variable("x"), Variable("y")), "x != y"),
    (GreaterThan(Variable("x"), Variable("y")), "x > y"),
    (LessThan(Variable("x"), Variable("y")), "x < y"),
    (GreaterThanOrEqual(Variable("x"), Variable("y")), "x >= y"),
    (LessThanOrEqual(Variable("x"), Variable("y")), "x <= y"),
    # Boolean operators
    (And(Variable("x"), Variable("y")), "x && y"),
    (Or(Variable("x"), Variable("y")), "x || y"),
    # Max/min
    (Max(Variable("x"), Variable("y")), "TACO_MAX(x, y)"),
    (Min(Variable("x"), Variable("y")), "TACO_MIN(x, y)"),
    (Max(Max(Variable("x"), Variable("y")), Variable("z")), "TACO_MAX(TACO_MAX(x, y), z)"),
    (Min(Min(Variable("x"), Variable("y")), Variable("z")), "TACO_MIN(TACO_MIN(x, y), z)"),
    # Cast
    (BooleanToInteger(Equal(Variable("x"), Variable("y"))), "(int32_t)(x == y)"),
    # Allocate
    (ArrayAllocate(integer, Variable("capacity")), "malloc(sizeof(int32_t) * capacity)"),
    (
        ArrayAllocate(float, Add(Variable("previous"), Variable("new"))),
        "malloc(sizeof(double) * (previous + new))",
    ),
    (
        ArrayReallocate(Variable("old"), integer, Variable("capacity")),
        "realloc(old, sizeof(int32_t) * capacity)",
    ),
    (
        ArrayReallocate(Variable("old"), float, Add(Variable("previous"), Variable("new"))),
        "realloc(old, sizeof(double) * (previous + new))",
    ),
    # Declaration and types
    (Declaration(Variable("x"), integer), "int32_t x"),
    (Declaration(Variable("x"), float), "double x"),
    (Declaration(Variable("x"), tensor), "taco_tensor_t x"),
    (Declaration(Variable("x"), Pointer(float)), "double* restrict x"),
    (Declaration(Variable("x"), Pointer(Pointer(integer))), "int32_t* restrict* restrict x"),
    (Declaration(Variable("x"), Array(float)), "double x[]"),
    (Declaration(Variable("x"), Array(Array(integer))), "int32_t x[][]"),
    (Declaration(Variable("x"), FixedArray(mode, 3)), "taco_mode_t x[3]"),
    (Declaration(Variable("x"), FixedArray(FixedArray(mode, 3), 2)), "taco_mode_t x[3][2]"),
    # Assignment
    (Assignment(Variable("x"), Variable("y")), "x = y"),
    (Assignment(Variable("x"), Add(Variable("x"), IntegerLiteral(1))), "x++"),
    (Assignment(Variable("x"), Add(Variable("x"), IntegerLiteral(2))), "x += 2"),
    (Assignment(Variable("x"), Subtract(Variable("x"), IntegerLiteral(1))), "x--"),
    (Assignment(Variable("x"), Subtract(Variable("x"), IntegerLiteral(2))), "x -= 2"),
    (Assignment(Variable("x"), Multiply(Variable("x"), IntegerLiteral(2))), "x *= 2"),
    (DeclarationAssignment(Declaration(Variable("x"), integer), Variable("y")), "int32_t x = y"),
    # Return
    (Return(IntegerLiteral(0)), "return 0"),
]


@pytest.mark.parametrize(("ast", "code"), single_lines)
def test_single_lines(ast: Expression, code: str):
    assert ir_to_c_statement(ast) == [code + ";"]


multiple_lines = [
    (
        Block(
            [
                DeclarationAssignment(Declaration(Variable("x"), integer), IntegerLiteral(0)),
                Assignment(Variable("x"), Add(Variable("x"), IntegerLiteral(2))),
            ]
        ),
        """
        int32_t x = 0;
        x += 2;
        """,
    ),
    (
        Block(
            [
                DeclarationAssignment(Declaration(Variable("x"), integer), IntegerLiteral(0)),
            ],
            "comment",
        ),
        """
        // comment
        int32_t x = 0;
        """,
    ),
    (
        Block(
            [
                Block(
                    [
                        DeclarationAssignment(
                            Declaration(Variable("x"), integer), IntegerLiteral(0)
                        ),
                    ],
                    "inner comment",
                ),
                DeclarationAssignment(Declaration(Variable("y"), integer), IntegerLiteral(0)),
            ],
        ),
        """
        // inner comment
        int32_t x = 0;

        int32_t y = 0;
        """,
    ),
    (
        Block(
            [
                DeclarationAssignment(Declaration(Variable("y"), integer), IntegerLiteral(0)),
                Block(
                    [
                        DeclarationAssignment(
                            Declaration(Variable("x"), integer), IntegerLiteral(0)
                        ),
                    ],
                    "inner comment",
                ),
            ],
        ),
        """
        int32_t y = 0;

        // inner comment
        int32_t x = 0;
        """,
    ),
    (
        Block(
            [
                Block(
                    [
                        DeclarationAssignment(
                            Declaration(Variable("x"), integer), IntegerLiteral(0)
                        ),
                    ],
                    "inner comment",
                ),
                DeclarationAssignment(Declaration(Variable("y"), integer), IntegerLiteral(0)),
                DeclarationAssignment(Declaration(Variable("z"), integer), IntegerLiteral(0)),
            ],
            "comment",
        ),
        """
        // comment

        // inner comment
        int32_t x = 0;

        int32_t y = 0;
        int32_t z = 0;
        """,
    ),
    (
        Branch(
            Equal(Variable("x"), Variable("y")),
            Assignment(Variable("z"), IntegerLiteral(0)),
            Assignment(Variable("z"), IntegerLiteral(1)),
        ),
        """
        if (x == y) {
          z = 0;
        } else {
          z = 1;
        }
        """,
    ),
    (
        Branch(
            Equal(Variable("x"), Variable("y")),
            Assignment(Variable("z"), IntegerLiteral(0)),
            Block([]),
        ),
        """
        if (x == y) {
          z = 0;
        }
        """,
    ),
    (
        Branch(
            Equal(Variable("x"), IntegerLiteral(0)),
            Assignment(Variable("y"), Variable("a")),
            Branch(
                Equal(Variable("x"), IntegerLiteral(1)),
                Assignment(Variable("y"), Variable("b")),
                Assignment(Variable("y"), Variable("c")),
            ),
        ),
        """
        if (x == 0) {
          y = a;
        } else if (x == 1) {
          y = b;
        } else {
          y = c;
        }
        """,
    ),
    (
        Branch(
            Equal(Variable("x"), IntegerLiteral(0)),
            Assignment(Variable("y"), Variable("a")),
            Branch(
                Equal(Variable("x"), IntegerLiteral(1)),
                Assignment(Variable("y"), Variable("b")),
                Branch(
                    Equal(Variable("x"), IntegerLiteral(2)),
                    Assignment(Variable("y"), Variable("c")),
                    Block([]),
                ),
            ),
        ),
        """
        if (x == 0) {
          y = a;
        } else if (x == 1) {
          y = b;
        } else if (x == 2) {
          y = c;
        }
        """,
    ),
    (
        Loop(
            LessThan(Variable("x"), Variable("y")),
            Assignment(Variable("x"), Add(Variable("x"), IntegerLiteral(1))),
        ),
        """
        while (x < y) {
          x++;
        }
        """,
    ),
    (
        Loop(
            LessThan(Variable("x"), Variable("y")),
            Block(
                [DeclarationAssignment(Declaration(Variable("x"), integer), IntegerLiteral(0))],
                "comment",
            ),
        ),
        """
        while (x < y) {
          // comment
          int32_t x = 0;
        }
        """,
    ),
]


@pytest.mark.parametrize(("ast", "code"), multiple_lines)
def test_multiple_lines(ast: Statement, code: str):
    assert "\n".join(ir_to_c_statement(ast)) == clean(code)


def test_function_definition():
    function = FunctionDefinition(
        Variable("double_x"),
        [Declaration(Variable("x"), integer)],
        integer,
        Return(Multiply(Variable("x"), IntegerLiteral(2))),
    )
    expected = """
        int32_t double_x(int32_t x) {
          return x * 2;
        }
        """
    assert ir_to_c_function_definition(function) == clean(expected)


def test_module():
    function = FunctionDefinition(
        Variable("double_x"),
        [Declaration(Variable("x"), integer)],
        integer,
        Return(Multiply(Variable("x"), IntegerLiteral(2))),
    )
    module = Module([function, function])
    expected = """
        int32_t double_x(int32_t x) {
          return x * 2;
        }

        int32_t double_x(int32_t x) {
          return x * 2;
        }
        """
    assert ir_to_c(module) == clean(expected)



__all__ = ["parse_assignment"]

from functools import reduce

from parsita import ParseError, ParserContext, lit, reg, rep, rep1sep, repsep
from parsita.util import splat
from returns import result

from ._exceptions import InconsistentDimensionsError, MutatingAssignmentError, NameConflictError
from .ast import Add, Assignment, Float, Integer, Multiply, Subtract, Tensor


def make_expression(first, rest):
    value = first
    for op, term in rest:
        match op:
            case "+":
                value = Add(value, term)
            case "-":
                value = Subtract(value, term)
    return value


class TensorExpressionParsers(ParserContext, whitespace=r"[ ]*"):
    name = reg(r"[A-Za-z][A-Za-z0-9]*")

    floating_point = reg(r"\d+((\.\d+([Ee][+-]?\d+)?)|((\.\d+)?[Ee][+-]?\d+))") > (
        lambda x: Float(float(x))
    )
    integer = reg(r"[0-9]+") > (lambda x: Integer(int(x)))
    number = floating_point | integer

    tensor = name & "(" >> (repsep(name, ",") > tuple) << ")" > splat(Tensor)

    parentheses = "(" >> expression << ")"  # noqa: F821
    factor = tensor | number | parentheses

    term = rep1sep(factor, "*") > (lambda x: reduce(Multiply, x))
    expression = term & rep(lit("+", "-") & term) > splat(make_expression)

    assignment = tensor & "=" >> expression > splat(Assignment)


def parse_assignment(
    string: str,
) -> result.Result[
    Assignment,
    ParseError | MutatingAssignmentError | InconsistentDimensionsError | NameConflictError,
]:
    try:
        return TensorExpressionParsers.assignment.parse(string)
    except (MutatingAssignmentError, InconsistentDimensionsError, NameConflictError) as e:
        return result.Failure(e)



__all__ = ["DiagonalAccessError", "NoKernelFoundError"]

from dataclasses import dataclass

from . import ast


@dataclass(frozen=True, slots=True)
class DiagonalAccessError(Exception):
    tensor: ast.Tensor

    def __str__(self) -> str:
        return (
            f"Diagonal access to a tensor (i.e. repeating the same index within a tensor) is not "
            f"currently supported: {self.tensor.name}({', '.join(self.tensor.indexes)})"
        )


@dataclass(frozen=True, slots=True)
class NoKernelFoundError(Exception):
    def __str__(self) -> str:
        return (
            "Tensora's tensor algebra compiler was unable to find a kernel for the given problem. "
            "This is likely due to sparse tensors needing to be iterated in opposite orders."
        )



from __future__ import annotations

__all__ = ["IterationGraph", "TerminalNode", "IterationNode", "SumNode"]

from abc import abstractmethod
from dataclasses import dataclass, replace

from .._stable_set import StableFrozenSet
from ..format import Mode
from .identifiable_expression import Context, TensorLayer, exhaust_tensor, extract_context
from .identifiable_expression.ast import Expression, Integer


class IterationGraph:
    @abstractmethod
    def extract_context(self, index: str) -> Context:
        raise NotImplementedError()

    @abstractmethod
    def exhaust_tensor(self, reference: str) -> IterationGraph:
        raise NotImplementedError()

    @abstractmethod
    def is_sparse_output(self) -> bool:
        # Needed by assembly to determine if next layer is guaranteed to advance position or not
        raise NotImplementedError()

    @abstractmethod
    def compressed_dimensions(self) -> StableFrozenSet[str]:
        # Needed when empty subgraphs simplify
        raise NotImplementedError()

    @abstractmethod
    def later_indexes(self) -> frozenset[str]:
        raise NotImplementedError()

    @abstractmethod
    def has_output(self) -> bool:
        raise NotImplementedError()


@dataclass(frozen=True)
class TerminalNode(IterationGraph):
    expression: Expression

    def extract_context(self, index: str) -> Context:
        return extract_context(self.expression, index)

    def exhaust_tensor(self, reference: str) -> IterationGraph:
        return TerminalNode(exhaust_tensor(self.expression, reference))

    def is_sparse_output(self) -> bool:
        return False

    def compressed_dimensions(self) -> StableFrozenSet[str]:
        # Needed when empty subgraphs simplify
        return StableFrozenSet()

    def later_indexes(self) -> frozenset[str]:
        return frozenset()

    def has_output(self) -> bool:
        return False


@dataclass(frozen=True)
class IterationNode(IterationGraph):
    index_variable: str
    output: TensorLayer | None
    next: IterationGraph

    def __post_init__(self):
        self.context: Context
        object.__setattr__(self, "context", self.extract_context(self.index_variable))

    def extract_context(self, index: str) -> Context:
        next_context = self.next.extract_context(index)
        return replace(
            next_context,
            indexes=next_context.indexes | frozenset([self.index_variable]),
            has_output=next_context.has_output or self.output is not None,
            has_assemble=next_context.has_assemble or self.is_sparse_output(),
        )

    def exhaust_tensor(self, reference: str) -> IterationGraph:
        new_next = self.next.exhaust_tensor(reference)

        return replace(self, next=new_next)

    def compressed_dimensions(self) -> StableFrozenSet[str]:
        return StableFrozenSet(*(leaf.tensor.id for leaf in self.context.sparse_leaves))

    def sparse_leaves(self) -> list[TensorLayer]:
        return [TensorLayer(leaf.tensor, leaf.layer) for leaf in self.context.sparse_leaves]

    def dense_leaves(self) -> list[TensorLayer]:
        return [TensorLayer(leaf.tensor, leaf.layer) for leaf in self.context.dense_leaves]

    def is_sparse_input(self) -> bool:
        return self.context.is_sparse

    def is_dense_output(self) -> bool:
        return self.output is not None and self.output.mode == Mode.dense

    def is_sparse_output(self) -> bool:
        return self.output is not None and self.output.mode == Mode.compressed

    def later_indexes(self) -> frozenset[str]:
        return self.context.indexes

    def has_output(self) -> bool:
        return self.context.has_output

    def has_assemble(self) -> bool:
        return self.context.has_assemble


@dataclass(frozen=True)
class SumNode(IterationGraph):
    name: str
    terms: list[IterationGraph]

    def extract_context(self, index: str) -> Context:
        context = Context(is_sparse=True)
        for term in self.terms:
            context = context.add(term.extract_context(index))
        return context

    def exhaust_tensor(self, reference: str) -> IterationGraph:
        new_terms = []
        for term in self.terms:
            new_term = term.exhaust_tensor(reference)
            # TODO: Simplify empty terms
            new_terms.append(new_term)

        if len(new_terms) == 0:
            return TerminalNode(Integer(0))
        elif len(new_terms) == 1:
            return new_terms[0]
        else:
            return replace(self, terms=new_terms)

    def is_sparse_output(self) -> bool:
        return False

    def compressed_dimensions(self) -> StableFrozenSet[str]:
        # Needed when empty subgraphs simplify
        return StableFrozenSet()

    def later_indexes(self) -> frozenset[str]:
        return frozenset.union(*(term.later_indexes() for term in self.terms))

    def has_output(self) -> bool:
        return any(term.has_output() for term in self.terms)



__all__ = ["TensorCompiler", "generate_code"]

from dataclasses import dataclass
from enum import Enum

from returns.result import Failure, Result, Success

from ..codegen import ir_to_c, ir_to_llvm
from ..kernel_type import KernelType
from ..problem import Problem
from ._taco import generate_c_code_taco
from ._tensora import generate_module_tensora


class TensorCompiler(str, Enum):
    """The tool to generate the tensor kernel.

    Attributes
    ----------
    tensora
        The tensor compiler included with Tensora.
    taco
        The original TACO binary repackaged in `tensora-taco`.
        Only available via the `taco` extra.
        Not compatible with `Language.llvm`.
        Not available on Windows.
    """

    # Python 3.10 does not support StrEnum, so do it manually
    tensora = "tensora"
    taco = "taco"

    def __str__(self) -> str:
        return self.name


class Language(str, Enum):
    """The language to be generated.

    Attributes
    ----------
    c
        C language.
    llvm
        LLVM IR.
        Not compatible with `TensorCompiler.taco`.
    """

    # Python 3.10 does not support StrEnum, so do it manually
    c = "c"
    llvm = "llvm"

    def __str__(self) -> str:
        return self.name


@dataclass(frozen=True, slots=True)
class UnsupportedLanguageError(Exception):
    tensor_compiler: TensorCompiler
    language: Language

    def __str__(self):
        return f"For tensor compiler {self.tensor_compiler}, language {self.language} is not supported."


def generate_code(
    problem: Problem,
    kernel_types: list[KernelType],
    tensor_compiler: TensorCompiler,
    language: Language,
) -> Result[str, Exception]:
    match tensor_compiler:
        case TensorCompiler.tensora:
            match generate_module_tensora(problem, kernel_types):
                case Failure(error):
                    return Failure(error)
                case Success(module):
                    match language:
                        case Language.c:
                            return Success(ir_to_c(module))
                        case Language.llvm:
                            return Success(str(ir_to_llvm(module)))
                case _:
                    raise NotImplementedError()
        case TensorCompiler.taco:
            match language:
                case Language.c:
                    return generate_c_code_taco(problem, kernel_types)
                case Language.llvm:
                    return Failure(UnsupportedLanguageError(tensor_compiler, language))



__all__ = ["BucketOutput"]

from dataclasses import dataclass, replace

from ...format import Mode
from ...ir import SourceBuilder, types
from ...ir.ast import Add, Expression, LessThan, Multiply, Variable
from ...kernel_type import KernelType
from .._names import dimension_name
from ..identifiable_expression import TensorLayer
from ..identifiable_expression import ast as ie_ast
from ._base import Output


@dataclass(frozen=True, slots=True)
class BucketOutput(Output):
    output: ie_ast.Tensor
    layers: list[int]
    unfulfilled: set[int]

    def __init__(
        self, output: ie_ast.Tensor, layers: list[int], unfulfilled: set[int] | None = None
    ):
        object.__setattr__(self, "output", output)
        object.__setattr__(self, "layers", layers)
        if unfulfilled is not None:
            object.__setattr__(self, "unfulfilled", unfulfilled)
        else:
            unfulfilled = {layer for layer in layers if output.modes[layer] == Mode.compressed}
            object.__setattr__(self, "unfulfilled", unfulfilled)

    def write_declarations(self, right_hand_side: Expression) -> SourceBuilder:
        source = SourceBuilder("Bucket initialization")
        source.append(self.name().declare(types.Pointer(types.float)).assign(right_hand_side))
        bucket_loop_index = self.loop_name()
        source.append(bucket_loop_index.declare(types.integer).assign(0))
        with source.loop(LessThan(bucket_loop_index, Multiply.join(self.dimension_names()))):
            source.append(self.name().idx(bucket_loop_index).assign(0))
            source.append(bucket_loop_index.increment())
        return source

    def next_output(
        self, iteration_output: TensorLayer | None, kernel_type: KernelType
    ) -> tuple[Output, SourceBuilder, SourceBuilder]:
        if iteration_output is None:
            return self, SourceBuilder(), SourceBuilder()
        else:
            next_unfulfilled = self.unfulfilled - {iteration_output.layer}
            return replace(self, unfulfilled=next_unfulfilled), SourceBuilder(), SourceBuilder()

    def write_assignment(self, right_hand_side: Expression, kernel_type: KernelType):
        source = SourceBuilder()
        bucket_index = self.ravel_indexes(
            self.dimension_names(),
            [Variable(self.output.indexes[layer]) for layer in self.layers],
        )
        source.append(self.name().idx(bucket_index).increment(right_hand_side))
        return source

    def ravel_indexes(self, dimensions: list[Variable], indexes: list[Variable]):
        dimensions_so_far: list[Variable] = []
        terms: list[Expression] = []
        for dim_i, index_i in zip(reversed(dimensions), reversed(indexes), strict=True):
            terms.append(Multiply.join([index_i, *dimensions_so_far]))
            dimensions_so_far.append(dim_i)

        # Reverse it to make it look nice
        return Add.join(list(reversed(terms)))

    def name(self) -> Variable:
        return Variable(f'bucket_{self.output.id}{"".join(f"_{x}" for x in self.layers)}')

    def loop_name(self) -> Variable:
        return Variable(f'i_bucket_{self.output.id}{"".join(f"_{x}" for x in self.layers)}')

    def dimension_names(self):
        return [dimension_name(self.output.indexes[layer]) for layer in self.layers]



import pytest

from tensora import Tensor
from tensora.compile import evaluate_cffi as evaluate


def assert_same_as_dense(expression, format_out, **tensor_pairs):
    tensors_in_format = {
        name: Tensor.from_lol(data, format=format) for name, (data, format) in tensor_pairs.items()
    }
    tensors_as_dense = {name: Tensor.from_lol(data) for name, (data, _) in tensor_pairs.items()}

    dense_format = "d" * (format_out.count("d") + format_out.count("s"))
    actual = evaluate(expression, format_out, **tensors_in_format)
    expected = evaluate(expression, dense_format, **tensors_as_dense)
    assert actual == expected


@pytest.mark.parametrize("dense", [[3, 2, 4], [0, 0, 0]])
@pytest.mark.parametrize("format_in", ["s", "d"])
@pytest.mark.parametrize("format_out", ["s", "d"])
def test_copy_1(dense, format_in, format_out):
    a = Tensor.from_lol(dense, format=format_in)
    actual = evaluate("b(i) = a(i)", format_out, a=a)
    assert actual == a


@pytest.mark.parametrize("dense", [[[0, 2, 4], [0, -1, 0]], [[0, 0, 0], [0, 0, 0]]])
@pytest.mark.parametrize("format_in", ["ss", "dd", "sd", "ds", "d1d0"])
@pytest.mark.parametrize("format_out", ["ss", "dd", "sd", "ds", "d1d0"])
def test_copy_2(dense, format_in, format_out):
    a = Tensor.from_lol(dense, format=format_in)
    actual = evaluate("b(i,j) = a(i,j)", format_out, a=a)
    assert actual == a


@pytest.mark.parametrize("dense", [[[0, 2, 4], [0, -1, 0]], [[0, 0, 0], [0, 0, 0]]])
@pytest.mark.parametrize("format_in", ["s1s0", "d1d0", "s1d0", "d1s0", "dd"])
@pytest.mark.parametrize("format_out", ["s1s0", "d1d0", "s1d0", "d1s0", "dd"])
def test_copy_2_backwards(dense, format_in, format_out):
    a = Tensor.from_lol(dense, format=format_in)
    actual = evaluate("b(i,j) = a(i,j)", format_out, a=a)
    assert actual == a


@pytest.mark.parametrize("expression", [0, 1])
def test_constant_scalar(expression):
    actual = evaluate(f"a() = {expression}", "")
    assert actual == Tensor.from_lol(expression)


@pytest.mark.parametrize("dense1", [[0, 2, 4, 0], [0, 0, 0, 0]])
@pytest.mark.parametrize("dense2", [[-1, 3.5, 0, 0], [0, 0, 0, 0]])
@pytest.mark.parametrize("format1", ["s", "d"])
@pytest.mark.parametrize("format2", ["s", "d"])
def test_vector_dot(dense1, dense2, format1, format2):
    assert_same_as_dense(
        "out() = in1(i) * in2(i)", "", in1=(dense1, format1), in2=(dense2, format2)
    )


@pytest.mark.parametrize("dense1", [[0, 2, 4, 0], [0, 0, 0, 0]])
@pytest.mark.parametrize("dense2", [[-1, 3.5, 0, 0], [0, 0, 0, 0]])
@pytest.mark.parametrize("format1", ["s", "d"])
@pytest.mark.parametrize("format2", ["s", "d"])
@pytest.mark.parametrize("format_out", ["s", "d"])
@pytest.mark.parametrize("operator", ["+", "-", "*"])
def test_vector_binary(operator, dense1, dense2, format1, format2, format_out):
    assert_same_as_dense(
        f"out(i) = in1(i) {operator} in2(i)",
        format_out,
        in1=(dense1, format1),
        in2=(dense2, format2),
    )


@pytest.mark.parametrize("dense1", [[[0, 2, 4], [0, -1, 0]], [[0, 0, 0], [0, 0, 0]]])
@pytest.mark.parametrize("dense2", [[[-1, 3.5], [0, 0], [4, 0]], [[0, 0], [0, 0], [0, 0]]])
@pytest.mark.parametrize("format1", ["ss", "dd", "sd", "ds", "d1d0"])
@pytest.mark.parametrize("format2", ["ss", "dd", "sd", "ds", "d1d0"])
@pytest.mark.parametrize("format_out", ["dd", "d1d0"])
def test_matrix_dot(dense1, dense2, format1, format2, format_out):
    assert_same_as_dense(
        "out(i,k) = in1(i,j) * in2(j,k)", format_out, in1=(dense1, format1), in2=(dense2, format2)
    )


@pytest.mark.parametrize("dense1", [[[0, 2, 4], [0, -1, 0]], [[0, 0, 0], [0, 0, 0]]])
@pytest.mark.parametrize("dense2", [[-1, 3.5, 0], [0, 0, 0]])
@pytest.mark.parametrize("format1", ["ss", "dd", "sd", "ds", "s1s0", "d1d0", "s1d0", "d1s0"])
@pytest.mark.parametrize("format2", ["s", "d"])
@pytest.mark.parametrize("format_out", ["d"])
def test_matrix_vector_product(dense1, dense2, format1, format2, format_out):
    assert_same_as_dense(
        "out(i) = in1(i,j) * in2(j)", format_out, in1=(dense1, format1), in2=(dense2, format2)
    )


@pytest.mark.parametrize("dense1", [[[0, 2, 4], [0, -1, 0]], [[0, 0, 0], [0, 0, 0]]])
@pytest.mark.parametrize("dense2", [[[-1, 3.5], [0, 0], [4, 0]], [[0, 0], [0, 0], [0, 0]]])
@pytest.mark.parametrize("dense3", [[[-3, 0], [7, 0]], [[0, 0], [0, 0]]])
@pytest.mark.parametrize("format1", ["dd", "ds"])
@pytest.mark.parametrize("format2", ["dd"])
@pytest.mark.parametrize("format3", ["dd", "ds"])
@pytest.mark.parametrize("format_out", ["dd"])
def test_matrix_multiply_add(dense1, dense2, dense3, format1, format2, format3, format_out):
    assert_same_as_dense(
        "out(i,k) = in1(i,j) * in2(j,k) + in3(i,k)",
        format_out,
        in1=(dense1, format1),
        in2=(dense2, format2),
        in3=(dense3, format3),
    )


@pytest.mark.parametrize(
    "dense_b",
    [
        [
            [[0, 2, 4, 0], [0, -1, 0, 3], [1, -1, 0, 0]],
            [[-2, 4, 0, 0], [0, 0, 0, 3], [1, 1, 0, 0]],
        ],
        [
            [[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]],
            [[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]],
        ],
    ],
)
@pytest.mark.parametrize(
    "dense_d",
    [
        [[-1, 3.5, 1, 2, 0], [0, 2, 6, 3, 0], [4, 0, 0, 1, -1], [0, 0, 3, 6, 9]],
        [[0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0]],
    ],
)
@pytest.mark.parametrize(
    "dense_c",
    [
        [[0, 0, 1, 2, 7], [7, 0, 5, 2, 0], [-1, 0, 0, 2, 1]],
        [[0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0]],
    ],
)
@pytest.mark.parametrize("format_b", ["ddd", "dss", "sss", "ssd", "d1d0s2", "s0d2d1"])
@pytest.mark.parametrize("format_d", ["dd", "ds", "ss"])
@pytest.mark.parametrize("format_c", ["dd", "ds", "ss"])
@pytest.mark.parametrize("format_out", ["dd", "d1d0"])
def test_mttkrp(dense_b, dense_d, dense_c, format_b, format_d, format_c, format_out):
    assert_same_as_dense(
        "A(i,j) = B(i,k,l) * D(l,j) * C(k,j)",
        format_out,
        B=(dense_b, format_b),
        D=(dense_d, format_d),
        C=(dense_c, format_c),
    )



import hypothesis.strategies as st
from hypothesis import assume

from tensora import Tensor
from tensora.expression import (
    InconsistentDimensionsError,
    MutatingAssignmentError,
    NameConflictError,
    ast,
)
from tensora.format import Format, Mode
from tensora.problem import Problem

names = st.from_regex(r"[A-Za-z][A-Za-z0-9]*", fullmatch=True)
variables = st.builds(
    ast.Tensor, name=names, indexes=st.builds(tuple, st.lists(names, max_size=4))
)
expressions = st.deferred(
    lambda: st.builds(ast.Integer, st.integers(min_value=0, max_value=2**16))
    | st.builds(ast.Float, st.floats(min_value=0, allow_infinity=False, allow_nan=False))
    | variables
    | adds
    | subtracts
    | multiplies
)
adds = st.builds(ast.Add, expressions, expressions)
subtracts = st.builds(ast.Subtract, expressions, expressions)
multiplies = st.builds(ast.Multiply, expressions, expressions)


@st.composite
def assignments(draw):
    target = draw(variables)
    expression = draw(expressions)
    try:
        return ast.Assignment(target, expression)
    except (InconsistentDimensionsError, MutatingAssignmentError, NameConflictError):
        assume(False)


modes = st.sampled_from(Mode)


@st.composite
def formats(draw, orders=st.integers(min_value=0, max_value=4)) -> Format:
    order = draw(orders)
    format_modes = tuple(draw(st.lists(modes, min_size=order, max_size=order)))
    format_mode_ordering = tuple(draw(st.permutations(range(order))))
    return Format(format_modes, format_mode_ordering)


@st.composite
def problems(draw) -> Problem:
    assignment = draw(assignments())

    problem_formats = {
        name: draw(formats(st.just(order))) for name, order in assignment.variable_orders().items()
    }

    return Problem(assignment, problem_formats)


@st.composite
def tensors(draw, format: Format, dimensions: tuple[int, ...] | None = None) -> Tensor:
    if dimensions is None:
        dimensions = draw(st.lists(st.integers(), min_size=format.order, max_size=format.order))

    if any(dim == 0 for dim in dimensions):
        # Hypothesis does not like being forced to draw empty dictionaries by giving it an
        # empty set of possible keys
        dok = {}
    else:
        dok = draw(
            st.dictionaries(
                st.tuples(*[st.integers(min_value=0, max_value=dim - 1) for dim in dimensions]),
                st.floats(allow_infinity=False, allow_nan=False),
            )
        )

    return Tensor.from_dok(dok, format=format, dimensions=dimensions)


@st.composite
def problem_and_tensors(draw):
    problem: Problem = draw(problems())

    # Indexes over the same dimension must have the same size.
    # Dimensions sharing an index must have the same size.
    participant_sizes: dict[tuple[str, int], int] = {}
    index_sizes: dict[str, int] = {}
    for index, participants in problem.assignment.index_participants().items():
        for participant in participants:
            if participant in participant_sizes:
                index_sizes[index] = participant_sizes[participant]
            elif index in index_sizes:
                participant_sizes[participant] = index_sizes[index]
            else:
                # This number can conspire with the number of dimensions to
                # require a LOT of memory.
                size = draw(st.integers(min_value=0, max_value=64))
                participant_sizes[participant] = size
                index_sizes[index] = size

    input_tensors = {}
    for name, variable in problem.assignment.expression.variables().items():
        format = problem.formats[name]
        dimensions = tuple(index_sizes[index] for index in variable[0].indexes)
        input_tensors[name] = draw(tensors(format, dimensions))

    return (problem, input_tensors)



__all__ = ["to_ir"]

from functools import singledispatch

from ...ir import ast as ir
from .ast import Add, Expression, Float, Integer, Multiply, Tensor


@singledispatch
def to_ir(self: Expression) -> ir.Expression:
    raise NotImplementedError(f"to_ir not implemented for {type(self)}: {self}")


@to_ir.register(Integer)
def to_ir_integer(self: Integer):
    # This is sensible as long as we only support floating point values and don't support division. If either of those
    # ceases to be true, this will need to be updated.
    return ir.IntegerLiteral(self.value)


@to_ir.register(Float)
def to_ir_float(self: Float):
    return ir.FloatLiteral(self.value)


@to_ir.register(Tensor)
def to_ir_tensor(self: Tensor):
    from .._names import previous_layer_pointer, vals_name

    return vals_name(self.name).idx(previous_layer_pointer(self.id, self.order))


@to_ir.register(Add)
def to_ir_add(self: Add):
    return ir.Add(to_ir(self.left), to_ir(self.right))


@to_ir.register(Multiply)
def to_ir_multiply(self: Multiply):
    return ir.Multiply(to_ir(self.left), to_ir(self.right))



from multiprocessing.pool import ThreadPool
from random import randrange

from tensora import Tensor, evaluate


def test_csr_matrix_vector_product():
    A = Tensor.from_aos([(1, 0), (0, 1), (1, 2)], [2.0, -2.0, 4.0], dimensions=(2, 3), format="ds")

    x = Tensor.from_aos([(0,), (1,), (2,)], [3.0, 2.5, 2.0], dimensions=(3,), format="d")

    expected = Tensor.from_aos([(0,), (1,)], [-5.0, 14.0], dimensions=(2,), format="d")

    actual = evaluate("y(i) = A(i,j) * x(j)", "d", A=A, x=x)

    assert actual == expected


def test_csc_matrix_vector_product():
    A = Tensor.from_aos(
        [(1, 0), (0, 1), (1, 2)], [2.0, -2.0, 4.0], dimensions=(2, 3), format="d1s0"
    )

    x = Tensor.from_aos([(0,), (1,), (2,)], [3.0, 2.5, 2.0], dimensions=(3,), format="d")

    expected = Tensor.from_aos([(0,), (1,)], [-5.0, 14.0], dimensions=(2,), format="d")

    actual = evaluate("y(i) = A(i,j) * x(j)", "d", A=A, x=x)

    assert actual == expected


def test_csr_matrix_plus_csr_matrix():
    A = Tensor.from_aos([(1, 0), (0, 1), (1, 2)], [2.0, -2.0, 4.0], dimensions=(2, 3), format="ds")

    B = Tensor.from_aos([(1, 1), (1, 2), (0, 2)], [-3.0, 4.0, 3.5], dimensions=(2, 3), format="ds")

    expected = Tensor.from_aos(
        [(1, 0), (0, 1), (1, 2), (1, 1), (0, 2)],
        [2.0, -2.0, 8.0, -3.0, 3.5],
        dimensions=(2, 3),
        format="ds",
    )

    actual = evaluate("C(i,j) = A(i,j) + B(i,j)", "ds", A=A, B=B)

    assert actual == expected


def test_rhs():
    A0 = Tensor.from_lol([2, -3, 0])
    A1 = Tensor.from_aos([(0, 2), (1, 2), (2, 2)], [3, 3, -3], dimensions=(3, 3), format="ds")
    A2 = Tensor.from_aos(
        [(0, 0, 1), (1, 0, 1), (2, 0, 1)], [-2, -2, 2], dimensions=(3, 3, 3), format="dss"
    )
    x = Tensor.from_lol([2, 3, 5])

    expected = Tensor.from_lol([5, 0, -3])

    assignment = "f(i) = A0(i) + A1(i,j) * x(j) + A2(i,k,l) * x(k) * x(l)"

    actual = evaluate(assignment, "d", A0=A0, A1=A1, A2=A2, x=x)

    assert actual == expected


def test_many_elements_stack_overflow():
    size = 1000000  # Big enough to trigger stack overflow

    a = Tensor.from_dok({}, dimensions=(size,), format="d")
    b = evaluate("b(i) = a(i)", "d", a=a)

    assert b == Tensor.from_dok({}, dimensions=(size,), format="d")


def test_many_elements_realloc():
    size = 2000000  # Big enough to trigger realloc

    a = Tensor.from_dok({}, dimensions=(size,), format="d")
    b = evaluate("b(i) = a(i)", "s", a=a)

    assert b == Tensor.from_dok({}, dimensions=(size,), format="s")


def test_multithread_evaluation():
    # As of version 1.14.4 of cffi, the FFI.compile method is not thread safe. This tests that evaluation of different
    # kernels is thread safe.
    A = Tensor.from_aos([(1, 0), (0, 1), (1, 2)], [2.0, -2.0, 4.0], dimensions=(2, 3), format="ds")

    x = Tensor.from_aos([(0,), (1,), (2,)], [3.0, 2.5, 2.0], dimensions=(3,), format="d")

    def run_eval():
        # Generate a random expression so that the cache cannot be hit
        return evaluate(f"y{randrange(1024)}(i) = A(i,j) * x(j)", "d", A=A, x=x)

    n = 4
    with ThreadPool(n) as p:
        results = p.starmap(run_eval, [()] * n)

    expected = run_eval()

    for actual in results:
        assert actual == expected



from __future__ import annotations

__all__ = [
    "Statement",
    "Expression",
    "Assignable",
    "Variable",
    "AttributeAccess",
    "ArrayIndex",
    "IntegerLiteral",
    "FloatLiteral",
    "BooleanLiteral",
    "Add",
    "Subtract",
    "Multiply",
    "Equal",
    "NotEqual",
    "GreaterThan",
    "GreaterThanOrEqual",
    "LessThan",
    "LessThanOrEqual",
    "And",
    "Or",
    "Max",
    "Min",
    "BooleanToInteger",
    "ArrayAllocate",
    "ArrayReallocate",
    "Declaration",
    "Assignment",
    "DeclarationAssignment",
    "Block",
    "Branch",
    "Loop",
    "Return",
    "FunctionDefinition",
    "Module",
]

from dataclasses import dataclass
from functools import reduce
from typing import Sequence

from .types import Type


def to_expression(expression: Expression | float | int | str) -> Expression:
    match expression:
        case Expression():
            return expression
        case float():
            return FloatLiteral(expression)
        case int():
            return IntegerLiteral(expression)
        case str():
            return Variable(expression)


class Statement:
    __slots__ = ()


class Expression(Statement):
    __slots__ = ()

    def plus(self, term: Expression | int | str) -> Expression:
        term = to_expression(term)
        return Add(self, term)

    def minus(self, term: Expression | int | str) -> Expression:
        term = to_expression(term)
        return Subtract(self, term)

    def times(self, factor: Expression | int | str) -> Expression:
        factor = to_expression(factor)
        return Multiply(self, factor)


class Assignable(Expression):
    __slots__ = ()

    def attr(self, attribute: str) -> Assignable:
        return AttributeAccess(self, attribute)

    def idx(self, index: Expression | int | str) -> Assignable:
        index = to_expression(index)
        return ArrayIndex(self, index)

    def assign(self, value: Expression | float | int | str) -> Statement:
        value = to_expression(value)
        return Assignment(self, value)

    def increment(self, amount: Expression | float | int | str = 1) -> Statement:
        amount = to_expression(amount)
        return self.assign(self.plus(amount))


@dataclass(frozen=True, slots=True)
class Variable(Assignable):
    name: str

    def declare(self, type: Type) -> Declaration:
        return Declaration(self, type)


@dataclass(frozen=True, slots=True)
class AttributeAccess(Assignable):
    # For languages that care, this represents attribute access to an object on
    # the heap. There is no way to use an object on the stack in this IR.
    target: Assignable
    attribute: str


@dataclass(frozen=True, slots=True)
class ArrayIndex(Assignable):
    # For languages that care, this represents attribute access to an array on
    # the heap. There is no way to use an array on the stack in this IR.
    target: Assignable
    index: Expression


@dataclass(frozen=True, slots=True)
class IntegerLiteral(Expression):
    value: int


@dataclass(frozen=True, slots=True)
class FloatLiteral(Expression):
    value: float


@dataclass(frozen=True, slots=True)
class BooleanLiteral(Expression):
    value: bool


@dataclass(frozen=True, slots=True)
class Add(Expression):
    left: Expression
    right: Expression

    @staticmethod
    def join(operands: Sequence[Expression | int | str]) -> Expression:
        expression_operands = [to_expression(operand) for operand in operands]
        return reduce(Add, expression_operands, IntegerLiteral(0))


@dataclass(frozen=True, slots=True)
class Subtract(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True, slots=True)
class Multiply(Expression):
    left: Expression
    right: Expression

    @staticmethod
    def join(operands: Sequence[Expression | int | str]) -> Expression:
        expression_operands = [to_expression(operand) for operand in operands]
        return reduce(Multiply, expression_operands, IntegerLiteral(1))


@dataclass(frozen=True, slots=True)
class Equal(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True, slots=True)
class NotEqual(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True, slots=True)
class GreaterThan(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True, slots=True)
class LessThan(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True, slots=True)
class GreaterThanOrEqual(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True, slots=True)
class LessThanOrEqual(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True, slots=True)
class And(Expression):
    left: Expression
    right: Expression

    @staticmethod
    def join(operands: Sequence[Expression | int | str]) -> Expression:
        expression_operands = [to_expression(operand) for operand in operands]
        return reduce(And, expression_operands, BooleanLiteral(True))


@dataclass(frozen=True, slots=True)
class Or(Expression):
    left: Expression
    right: Expression

    @staticmethod
    def join(operands: Sequence[Expression | int | str]) -> Expression:
        expression_operands = [to_expression(operand) for operand in operands]
        return reduce(Or, expression_operands, BooleanLiteral(False))


@dataclass(frozen=True, slots=True)
class Max(Expression):
    left: Expression
    right: Expression

    @staticmethod
    def join(operands: Sequence[Expression | int | str]) -> Expression:
        expression_operands = [to_expression(operand) for operand in operands]
        return reduce(Max, expression_operands)


@dataclass(frozen=True, slots=True)
class Min(Expression):
    left: Expression
    right: Expression

    @staticmethod
    def join(operands: Sequence[Expression | int | str]) -> Expression:
        expression_operands = [to_expression(operand) for operand in operands]
        return reduce(Min, expression_operands)


@dataclass(frozen=True, slots=True)
class BooleanToInteger(Expression):
    expression: Expression


@dataclass(frozen=True, slots=True)
class ArrayAllocate(Expression):
    element_type: Type
    n_elements: Expression


@dataclass(frozen=True, slots=True)
class ArrayReallocate(Expression):
    old: Assignable
    element_type: Type
    n_elements: Expression


@dataclass(frozen=True, slots=True)
class Declaration(Statement):
    name: Variable
    type: Type

    def assign(self, value: Expression | int | str) -> Statement:
        value = to_expression(value)
        return DeclarationAssignment(self, value)


@dataclass(frozen=True, slots=True)
class Assignment(Statement):
    target: Assignable
    value: Expression


@dataclass(frozen=True, slots=True)
class DeclarationAssignment(Statement):
    target: Declaration
    value: Expression


@dataclass(frozen=True, slots=True)
class Block(Statement):
    statements: list[Statement]
    comment: str | None = None

    def is_empty(self):
        return len(self.statements) == 0


@dataclass(frozen=True, slots=True)
class Branch(Statement):
    condition: Expression
    if_true: Statement
    if_false: Statement

    @staticmethod
    def join(leaves: Sequence[tuple[Expression | int | str, Statement]]) -> Statement:
        # This is a fold right operation
        return reduce(
            lambda previous, leaf: Branch(to_expression(leaf[0]), leaf[1], previous),
            reversed(leaves),
            Block([]),
        )


@dataclass(frozen=True, slots=True)
class Loop(Statement):
    condition: Expression
    body: Statement


@dataclass(frozen=True, slots=True)
class Return(Statement):
    value: Expression


@dataclass(frozen=True, slots=True)
class FunctionDefinition:
    name: Variable
    parameters: list[Declaration]
    return_type: Type
    body: Statement


@dataclass(frozen=True, slots=True)
class Module:
    definitions: list[FunctionDefinition]
